

import datetime
import os
import struct

from constants import *
from utils.appdirs import user_data_dir
from utils.sdk_utils import (print_and_log, ERROR, DEBUG, bytearray_to_image,
                            db_person_to_myme_person, db_face_to_myme_face)

import threading
from sqlalchemy import (Column, ForeignKey, Integer, String, LargeBinary,
                        Boolean, DateTime, and_, create_engine)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, backref,sessionmaker, scoped_session
from sqlalchemy.pool import StaticPool

Base = declarative_base()

class DbFace(Base):
    __tablename__ = 'face'
    id = Column(type_=Integer, primary_key=True)
    myme_id = Column(type_=Integer, nullable=True) # Id of the face image internal to MyMe application
    detected_time = Column(type_=DateTime, nullable=False,
                           default=datetime.datetime.now())
    signature = Column(type_=LargeBinary, nullable=False)
    image = Column(type_=LargeBinary)
    image_fetch_failed = Column(type_=Boolean, nullable=False, default=False)
    person_id = Column(ForeignKey(column='person.id'))

    # this defines a one-to-many relation between the DbPerson and the DbFace objects
    # the cascade argumnet means every db operation is cascaded, so deleting a face from a
    # person object, would also delete it from the faces table, etc.
    # delete-orphan means un-refereced objects are deleted (like a face with no persons)
    # see more at http://docs.sqlalchemy.org/en/latest/orm/basic_relationships.html#one-to-many
    person = relationship("DbPerson",
                        backref=backref("faces", cascade="all, delete-orphan"))

    def __init__(self, myme_id=None, detected_time=None, signature=None, image=None,
                 image_fetch_failed=None, person_id=None):
        self.myme_id=myme_id
        if type(detected_time) is not datetime.datetime:
            self.detected_time = datetime.datetime.utcfromtimestamp(detected_time)
        else:
            self.detected_time = detected_time


        if signature is not None:
            self.signature = struct.pack('<{}b'.format(len(signature)), *signature)
        else:
            self.signature = None
        self.image=image
        self.image_fetch_failed=image_fetch_failed
        self.person_id=person_id


    def get_signature(self):
        if self.signature is None:
            return None
        else:
            return struct.unpack('<{}b'.format(len(self.signature)), str(self.signature))

    def get_image(self):
        if self.image is None:
            return None
        else:
            return bytearray_to_image(self.image)


class DbPerson(Base):
    __tablename__ = 'person'

    id = Column(Integer, primary_key=True)
    name = Column(String(250))
    profile_pic = Column(LargeBinary)

class DatabaseManager(object):
    MYME_DB_PATH = os.path.join(
            user_data_dir(appname=SDK_APPNAME, version=SDK_VERSION),
            MYME_DB_NAME)

    def __init__(self):

        print_and_log('database path {} '.
                           format(self.MYME_DB_PATH), DEBUG)
        self.lock = threading.Lock()

        db_url = "sqlite:///" + self.MYME_DB_PATH
        # Create an engine that stores data local folder (MYME_DB_PATH)
        # Note this can be changed to access other databases, also web databases
        if not os.path.exists(os.path.dirname(self.MYME_DB_PATH)):
            os.makedirs(os.path.dirname(self.MYME_DB_PATH))
        self.engine = create_engine(db_url, connect_args={'check_same_thread':False}, poolclass=StaticPool)

        # make sure tables are created
        self._create_tables()

        # Bind the engine to the metadata of the Base class so that the
        # declaratives can be accessed through a DBSession instance
        Base.metadata.bind = self.engine

        self.session_maker = sessionmaker(bind=self.engine)
        # Create a thread-local session
        # The ScopedSession object by default uses [threading.local()] as storage,
        # so that a single Session is maintained for all who call upon the
        # ScopedSession registry, but only within the scope of a single thread.
        # Callers who call upon the registry in a different thread get a Session
        # instance that is local to that other thread.
        # Using this technique, the ScopedSession provides a quick and relatively
        # simple way of providing a single, global object in an application that
        # is safe to be called upon from multiple threads.
        # TL;DR: SINCE DB IS USED ACROSS DIFFERENT THREADS, MUST USE SCOPED_SESSION

        # A DBSession() instance establishes all conversations with the database
        # and represents a "staging zone" for all the objects loaded into the
        # database session object. Any change made against the objects in the
        # session won't be persisted into the database until you call
        # session().commit(). If you're not happy about the changes, you can
        # revert all of them back to the last commit by calling
        # session().rollback()

        self.session = scoped_session(self.session_maker)

    def add(self, db_obj):
        with self.lock:
            self.session().add(db_obj)

    def refresh(self, db_obj):
        with self.lock:
            self.session().flush()
            self.session().refresh(db_obj)

    def add_and_refresh(self, db_obj):
        self.add(db_obj)
        self.refresh(db_obj)

    def add_and_commit(self, db_obj):
        with self.lock:
            self.session().add(db_obj)
            self.session().commit()

    def remove(self, db_obj):
        with self.lock:
            self.session().delete(db_obj)
            self.session().commit()

    def get_all_people(self):
        with self.lock:
            return self.session().query(DbPerson).all()

    def get_all_faces(self):
        with self.lock:
            return self.session().query(DbFace).all()

    def get_face_by_face_id(self, db_face_id):
        with self.lock:
            return self.session().query(DbFace).get(db_face_id)


    def get_person_by_person_id(self, person_id):
        with self.lock:
            return self.session().query(DbPerson).get(person_id)

    def remove_person_by_person_id(self, person_id):
        with self.lock:
            self.remove(self.get_person_by_person_id(person_id))
            self.session().commit()


    def get_all_faces_from_time(self, start_time, end_time):
        with self.lock:
            db_faces = self.session().query(DbFace).filter(
                    and_(DbFace.detected_time >  start_time,
                    DbFace.detected_time <  end_time)
            )
            return db_faces


    def add_face_image_to_id(self, db_faceid, image_bytearray):
        db_face = self.get_face_by_face_id(db_faceid)
        if db_face is None:
            print_and_log('ERROR: person id {0} does not exist'.
                           format(db_faceid), ERROR)
            return
        db_face.image = image_bytearray
        self.commit_db()

    def commit_db(self):
        """
        updates the DB with any pending changes (changes performed on any of
        the DbFaces/DbPerson instances)
        """
        with self.lock:
            self.session().commit()

    def _create_tables(self):
         # Create all tables in the engine (if not already exisiting)
        Base.metadata.create_all(self.engine)

    def _delete_db(self):
        """
        deletes all the tables in the database(!).
        :return:
        """
        Base.metadata.drop_all()
        self._create_tables()



import logging

from ... import BLEDevice, exceptions

log = logging.getLogger(__name__)


def connection_required(func):
    """Raise an exception before calling the actual function if the device is
    not connection.
    """
    def wrapper(self, *args, **kwargs):
        if not self._connected:
            raise exceptions.NotConnectedError()
        return func(self, *args, **kwargs)
    return wrapper


class GATTToolBLEDevice(BLEDevice):
    """A BLE device connection initiated by the GATTToolBackend.

    Since the GATTToolBackend can only support 1 device connection at at time,
    the device implementation defers to the backend for all functionality -
    every command has to synchronize around a the same interactive gatttool
    session, using the same connection.
    """
    def __init__(self, address, backend):
        super(GATTToolBLEDevice, self).__init__(address)
        self._backend = backend
        self._connected = True

    @connection_required
    def bond(self, *args, **kwargs):
        self._backend.bond(self, *args, **kwargs)

    @connection_required
    def char_read(self, uuid, *args, **kwargs):
        return self._backend.char_read(self, uuid, *args, **kwargs)

    @connection_required
    def char_write_handle(self, handle, *args, **kwargs):
        self._backend.char_write_handle(self, handle, *args, **kwargs)

    @connection_required
    def disconnect(self):
        self._backend.disconnect(self)
        self._connected = False

    @connection_required
    def discover_characteristics(self):
        return self._backend.discover_characteristics(self)



from ...exceptions import BLEError


class BGAPIError(BLEError):
    pass


class ExpectedResponseTimeout(BGAPIError):
    def __init__(self, expected_packets, timeout):
        super(ExpectedResponseTimeout, self).__init__(
            "Timed out after %fs waiting for %s" % (
                timeout or 0, expected_packets))




import abc

from enum import Enum

class UserCallbackReqType(Enum):

    # These cb requests call a delegate callback
    PairedDeviceStateUpdatedCbReq = 1
    FaceDetectedCbReq = 2
    NewPersonCreatedCbReq = 3
    VisualTopicsReceivedCbReq = 4
    OcrLineFoundCbReq = 5
    BarcodeDetectedCbReq = 6
    ChargerStateUpdatedCbReq = 7
    DeviceTappedCbReq = 8
    BatteryStateUpdatedCbReq = 9
    BatteryChargeLevelChangedCbReq = 10

    # These requests must get user_callback argument to call
    ScanCompleteCbReq = 11
    ConnectedToDeviceCbReq = 12
    FaceImageReceivedCbReq = 13
    FullImageReceivedCbReq = 14
    WriteToDebugCbReq = 15

class UserCallbackReq(object):
    __metaclass__ = abc.ABCMeta
    def __init__(self):
        self.type = None

class PairedDeviceStateUpdatedCbReq(UserCallbackReq):
    def __init__(self, new_state):
        """
        This delegate is invoked when the MyMe device state changes
        :param new_state: - New device state
        """
        UserCallbackReq.__init__(self)
        self.type = UserCallbackReqType.PairedDeviceStateUpdatedCbReq
        self.new_state = new_state


class FaceDetectedCbReq(UserCallbackReq):
    def __init__(self, face, person, probability, is_last_face_in_frame):
        """
        This delegate is invoked when the MyMe device detects a face
        :param face: MyMeFace object holding the matched face does not include
                     the actual image nor image signature.
        :param person: MyMePerson object corresponding to the matched person
        :param probability: probability that <face> belongs to <person>
        """
        UserCallbackReq.__init__(self)
        self.type = UserCallbackReqType.FaceDetectedCbReq
        self.face=face
        self.person=person
        self.probability=probability
        self.is_last_face_in_frame=is_last_face_in_frame


class NewPersonCreatedCbReq(UserCallbackReq):
    def __init__(self, person, timestamp):
        """
        This delegate is invoked when the MyMe device detects multiple faces
        and recognizes all of them as the same person. The DB creates a new
        person with the faces.
        :param person: newly created person (MyMePerson object)
        :param timestamp: time when face was created (seconds from epoch float)
        """
        UserCallbackReq.__init__(self)
        self.type = UserCallbackReqType.NewPersonCreatedCbReq
        self.person=person
        self.timestamp=timestamp


class VisualTopicsReceivedCbReq(UserCallbackReq):
    def __init__(self, topics, timestamp):
        """
        This delegate is invoked when a new set of words is sent from the device.
        :param topics: list containing objects of type VisualTopic which
        contains topic and its probability of being in the frame. The list is
        sorted by probability.
        :param timestamp: the time that the words were sent from the device
        """
        UserCallbackReq.__init__(self)
        self.type = UserCallbackReqType.VisualTopicsReceivedCbReq
        self.topics=topics
        self.timestamp=timestamp


class OcrLineFoundCbReq(UserCallbackReq):
    def __init__(self,  myme_ocr_line, timestamp, is_frame_end=False):
        """
        Inovked when a line of text is detected in an image.
        :param myme_ocr_line: Line of text detected (OcrLine object)
        :param timestamp: Timestamp when line was detected
        :param is_frame_end: if true, this OCR packet is the last one
                             for the current frame
                            (packet may be empty or contain OCR line(s))
        """
        UserCallbackReq.__init__(self)
        self.type = UserCallbackReqType.OcrLineFoundCbReq
        self.myme_ocr_line=myme_ocr_line
        self.timestamp=timestamp
        self.is_frame_end=is_frame_end


class BarcodeDetectedCbReq(UserCallbackReq):
    def __init__(self, barcode, timestamp):
        """
        Invoked when a product barcode is detected in an image.
        :param barcode: The detected barcode (ascii string)
        :param timestamp: Time when barcode was detected
        """
        UserCallbackReq.__init__(self)
        self.type = UserCallbackReqType.BarcodeDetectedCbReq
        self.barcode = barcode
        self.timestamp = timestamp


class ChargerStateUpdatedCbReq(UserCallbackReq):
    def __init__(self, new_charger_state):
        """
        This delegate is invoked when the charger state is changed.
        :param new_charger_state: the state of the charger after the change
        """
        UserCallbackReq.__init__(self)
        self.type = UserCallbackReqType.ChargerStateUpdatedCbReq
        self.new_charger_state = new_charger_state


class DeviceTappedCbReq(UserCallbackReq):
    def __init__(self):
        """
        This delegate is called when the device is tapped
        """
        UserCallbackReq.__init__(self)
        self.type = UserCallbackReqType.DeviceTappedCbReq


class BatteryStateUpdatedCbReq(UserCallbackReq):
    def __init__(self, battery_state):
        """
        Called when the battery state has been updated
        :param battery_state: BatteryState enum with the following options:
                            BatteryFull = 0
                            BatteryOK = 1
                            BatteryLow = 2
                            BatteryCritical = 3
        """
        UserCallbackReq.__init__(self)
        self.type = UserCallbackReqType.BatteryStateUpdatedCbReq
        self.battery_state = battery_state


class BatteryChargeLevelChangedCbReq(UserCallbackReq):
    def __init__(self, battery_level):
        """
        Called when the battery level is changed (expected to be called about
        once a minute while the device is connected)
        :param battery_level: Int from 1 to 100 (100 means charged 100%)
        """
        UserCallbackReq.__init__(self)
        self.type = UserCallbackReqType.BatteryChargeLevelChangedCbReq
        self.battery_level = battery_level


class ScanCompleteCbReq(UserCallbackReq):
    def __init__(self, sorted_device_list, scan_state, error, user_callback):
        """
        Called when a scan (following a call to scan_for_devices_with_timeout)
        is completed.
        :param sorted_device_list: List of devices.
        Each device is represented by dict with the following keys:
            "address" (device's MAC address)
            "name" (device name)
            "rssi" (signal strength)
        :param scan_state: scan compeltion reason (ScanState)
        :param error: in case of error, contains error message
        :param user_callback: User callback function to call
        """
        UserCallbackReq.__init__(self)
        self.type = UserCallbackReqType.ScanCompleteCbReq
        self.sorted_device_list = sorted_device_list
        self.scan_state = scan_state
        self.error = error
        self.user_callback = user_callback


class ConnectedToDeviceCbReq(UserCallbackReq):
    def __init__(self, device_addr, error, user_callback):
        """
        Iinvoked after an attempt is made to connect to a MyMe
        device. On success, device_addr and device_name are the address/name
        of the MyMe device that user connected to. On failure, error contains
        the error message.
        :param device_addr: address of the device that user connected to
        :param error: on connection failure, contains error string
        :param user_callback: User callback function to call

        """
        UserCallbackReq.__init__(self)
        self.type = UserCallbackReqType.ConnectedToDeviceCbReq
        self.device_addr = device_addr
        self.error = error
        self.user_callback = user_callback


class FaceImageReceivedCbReq(UserCallbackReq):
    def __init__(self, myme_face, user_callback):
        """
        :param myme_face: MyMeFace object containing the received face image
        :param user_callback: User callback function to call
        """
        UserCallbackReq.__init__(self)
        self.type = UserCallbackReqType.FaceImageReceivedCbReq
        self.myme_face = myme_face
        self.user_callback = user_callback


class FullImageReceivedCbReq(UserCallbackReq):
    def __init__(self, myme_image, user_callback):
        """
        :param MyMeImage object containing the received full image
        :param user_callback: User callback function to call
        """
        UserCallbackReq.__init__(self)
        self.type = UserCallbackReqType.FullImageReceivedCbReq
        self.myme_image = myme_image
        self.user_callback = user_callback

class WriteToDebugCbReq(UserCallbackReq):
    def __init__(self, success, user_callback):
        """
        :param success: True if succeeded in writing message to MyMe device
        :param user_callback: User callback function to call
        :return:
        """
        UserCallbackReq.__init__(self)
        self.type = UserCallbackReqType.WriteToDebugCbReq
        self.success = success
        self.user_callback = user_callback



import struct
from ..utils.sdk_utils import print_and_log, ERROR, EXCEPTION

PACKET_MIN_SIZE = 1 + 4

class PacketAssembler(object):
    """
    This class assembels packets sent from the MyMe device into one message
    """
    def __init__(self, callback):
        """
         :param callback - the callback to be called upon packet completion
         :return:
         """
        self.callback = callback
        self.current_data = bytearray()
        self.total_expeceted_len = 0

    def reset(self):
        """
        resets the state of the PacketAssembler, removing any current stored data
         and allowing new data to be sent.

        :return:
        """
        self.total_expeceted_len = 0
        self.current_data = bytearray()

    def send(self, packet):
        """
        sends a anew packet to be assmebled.
        Call this method for each new packet.
        Once fully assembled the calllback given on init would be called
        After that, new calls to send might be done (without need to call the 'reset' API).
        :return:
        """
        if self.total_expeceted_len == 0:
            if len(packet) < PACKET_MIN_SIZE:
                print_and_log('got packet of wrong size {0}'.format(len(packet)), ERROR)
                return

            # first packet, parse the header to get the size
            # < - little endian, x-padding, I - unsigned byte
            self.total_expeceted_len = struct.unpack('<xI', str(packet[0:5]))[0]

        self.current_data += (packet)

        # check if we've reached the end
        if len(self.current_data) >= self.total_expeceted_len:

            # the data is not expeceted to be ever bigger than the total expected
            if len(self.current_data) > self.total_expeceted_len:
                print_and_log('total size of data {0} is bigger than expected '
                    '{1}!'.format(self.current_data, self.total_expeceted_len),
                              ERROR)

            # time to call the callback
            self.callback(self.current_data)

            #  and some clean-up
            self.reset()



ble_address_type = {
    'gap_address_type_public': 0,
    'gap_address_type_random': 1
}
gap_discoverable_mode = {
    'non_discoverable': 0x00,
    'limited_discoverable': 0x01,
    'general_discoverable': 0x02,
    'broadcast': 0x03,
    'user_data': 0x04,
    'enhanced_broadcasting': 0x80
}
gap_connectable_mode = {
    'non_connectable': 0x00,
    'directed_connectable': 0x01,
    'undirected_connectable': 0x02,
    'scannable_non_connectable': 0x03,
}
gap_discover_mode = {
    'limited': 0x00,
    'generic': 0x01,
    'observation': 0x02,
}
bonding = {  # create bonding if devices not already bonded
    'do_not_create_bonding': 0x00,
    'create_bonding': 0x01,
}
bondable = {
    'no': 0x00,
    'yes': 0x01,
}
connection_status_flag = {
    'connected': 0x01,
    'encrypted': 0x02,
    'completed': 0x04,
    'parameters_change': 0x08,
}
scan_response_packet_type = {
    0x00: 'connectable_advertisement_packet',
    0x02: 'non-connectable_advertisement_packet',
    0x04: 'scan_response_packet',
    0x06: 'discoverable_advertisement_packet',
}
scan_response_data_type = {
    0x01: 'flags',
    0x02: 'incomplete_list_16-bit_service_class_uuids',
    0x03: 'complete_list_16-bit_service_class_uuids',
    0x04: 'incomplete_list_32-bit_service_class_uuids',
    0x05: 'complete_list_32-bit_service_class_uuids',
    0x06: 'incomplete_list_128-bit_service_class_uuids',
    0x07: 'complete_list_128-bit_service_class_uuids',
    0x08: 'shortened_local_name',
    0x09: 'complete_local_name',
    0x0A: 'tx_power_level',
    0x0D: 'class_of_device',
    0x0E: 'simple_pairing_hash_c/c-192',
    0x0F: 'simple_pairing_randomizer_r/r-192',
    0x10: 'device_id/security_manager_tk_value',
    0x11: 'security_manager_out_of_band_flags',
    0x12: 'slave_connection_interval_range',
    0x14: 'list_of_16-bit_service_solicitation_uuids',
    0x1F: 'list_of_32-bit_service_solicitation_uuids',
    0x15: 'list_of_128-bit_service_solicitation_uuids',
    0x16: 'service_data/service_data-16-bit_uuid',
    0x20: 'service_data-32-bit_uuid',
    0x21: 'service_data-128-bit_uuid',
    0x22: 'LE_secure_connections_confirmation_value',
    0x23: 'LE_secure_connections_random_value',
    0x17: 'public_target_address',
    0x18: 'random_target_address',
    0x19: 'appearance',
    0x1A: 'advertising_interval',
    0x1B: 'LE_bluetooth_device_address',
    0x1C: 'LE_role',
    0x1D: 'simple_pairing_hash_c-256',
    0x1E: 'simple_pairing_randomizer_r-256',
    0x3D: '3D_information_data',
    0xFF: 'manufacturer_specific_data',
}

# GATT
gatt_service_uuid = {
    'generic_access_profile': bytearray([0x18, 0x00]),
    'generic_attribute_profile': bytearray([0x18, 0x01]),
    'battery_service':bytearray([0x18, 0x0F]),
    'device_information':bytearray([0x18, 0x0A]),
}
gatt_attribute_type_uuid = {
    'primary_service': bytearray([0x28, 0x00]),
    'secondary_service': bytearray([0x28, 0x01]),
    'include': bytearray([0x28, 0x02]),
    'characteristic': bytearray([0x28, 0x03]),
}
gatt_characteristic_descriptor_uuid = {
    'characteristic_extended_properties': bytearray([0x29, 0x00]),
    'characteristic_user_description': bytearray([0x29, 0x01]),
    'client_characteristic_configuration': bytearray([0x29, 0x02]),
    'server_characteristic_configuration': bytearray([0x29, 0x03]),
    'characteristic_format': bytearray([0x29, 0x04]),
    'characteristic_aggregate_format': bytearray([0x29, 0x05]),
}
gatt_characteristic_type_uuid = {
    'device_name': bytearray([0x2A, 0x00]),
    'appearance': bytearray([0x2A, 0x01]),
    'peripheral_privacy_flag': bytearray([0x2A, 0x02]),
    'reconnection_address': bytearray([0x2A, 0x03]),
    'peripheral_preferred_connection_parameters': bytearray([0x2A, 0x04]),
    'service_changed': bytearray([0x2A, 0x05]),
    'battery_level': bytearray([0x2A, 0x19]),
    'hardware_revision_string': bytearray([0x2A, 0x27]),
    'software_revision_string': bytearray([0x2A, 0x28])
}




#from scipy.spatial.distance import cdist

import numpy as np
import time

from ..database_manager import DbPerson, DbFace
from ..myme_data_objects import MyMeFace

class FaceSession(object):
    MIN_SIZE_TO_CREATE_NEW_PERSON = 3
    DETECTOR_THRESHOLD=6.5
    CLARITY_THRESHOLD=22.5
    THRESHOLD_FACE_SESSION=25.65
    SESSION_TO_PERSON_THRESHOLD=27.3
    SIGNATURE_LEN = 128.

    def __init__(self, db_manager):
        self.db_manager = db_manager
        self.db_face_ids = [] # List of faces id in current session
        self.db_person_id = None # dbPerson id that most closely matches the session's sigs
        self.db_person_prob = 0 # Probability that this session matches <person>
        self.session_sigs_2d_nparray = np.empty((0, self.SIGNATURE_LEN), np.uint8)

        # Session is closed when 60 seconds pass since last face is added
        self.last_face_added_timestamp = time.time()

    def add_face(self, myme_face_id, detected_time, signature):
        """
        Adds a new face to the current session. Additionally, the face is
        added to the databse, and the DbFace object is refreshed.
        This is done so that the database always contains all currently active
        faces. This is useful in case where the user requests to get all faces.
        Also, by adding the face to the databse, the face gets a unique id
        which the user can then use to fetch an image for the face
        (the user will use the database id to access the face, while internally
        the SDK will use the myme_id to request an image from MyMe application)
        :param myme_face_id: Id of the face internal to MyMe application
                             will be stored in the database to enable later
                             fetching of image for the face
        :param detected_time: Time when face was deteced by MyMe
        :param signature: Signature of the detected faces (array of 128 ints)
        :return MyMeFace object representing the newly added face
        """
        db_face = DbFace(myme_id=myme_face_id,
                         detected_time=detected_time,
                         signature=signature)
        self.db_manager.add_and_commit(db_face)
        self.db_face_ids.append(db_face.id)
        self.last_face_added_timestamp = time.time()
        self.session_sigs_2d_nparray = \
            np.vstack([self.session_sigs_2d_nparray, np.array(signature)])

        return MyMeFace(id=db_face.id, detected_time=detected_time,
                        signature=signature)

    def close(self):
        """
        Close the session.
        On session closure one of two things can happen:
            1. Ignore (less than 3 face signatures in the session)
            2. Session faces are added to the DB and db_person is updated with
               all the new faces
        """
        print 'closing session'

        # get a list of db faces objects using the ids list
        db_faces = [self.db_manager.get_face_by_face_id(face_id) for face_id in self.db_face_ids]

        if len(self.db_face_ids) < self.MIN_SIZE_TO_CREATE_NEW_PERSON:
            # Remove the faces
            for face in db_faces:
                self.db_manager.remove(face)
            self.db_manager.commit_db()
            return


        if self.db_person_id is not None:
            for db_face in db_faces:
                db_face.person_id = self.db_person_id

        self.db_manager.commit_db()


    def create_person_and_close(self):
        # Create new person if no person has been matched to session faces
        db_person = DbPerson()
        self.db_manager.add_and_commit(db_person)
        self.db_person_id = db_person.id
        # Person created specifically for current faces to probability of match
        # is trivially 100%
        self.db_probability = 100
        self.close()


    def sig_list_to_2d_numpy_array(self, siglist):
        '''
        Given a regular list k of signatures of from:
        [ [a1,..,a128],...,[k1,...,k128] ]

        1. Converts it to a list of numpy arrays:
        [ np.array([a1,a2,...,a128]), ,..... ,np.array([z1,z2,....,z128]) ]

        2. And then creates and returns a 2D numpy array of signatures

        The returned matrix is of dimension N x 128 (where N is the number of
        signatures in the current session)
        np.array([a1,a2,..,a128],
                 [b1,b2,..,b128],
                 .
                 .
                 [k1,k2,..,k128])
        :param siglist:
        :return: 2D numpy array of signatures
        '''
        # Convert siglist to a list of numpy arrays:
        sigs_np_array = [np.array(sig) for sig in siglist]

        # Create a 2D numpy array of signatures
        sigs_2d_np_array = \
                np.concatenate([sig[np.newaxis, :] for sig in sigs_np_array])
        return sigs_2d_np_array


    def compare_session_sigs_to_person_sigs(self, session_sigs_2d_np_array,
                                                  person_sigs_np_array):
        """
        :param session_sigs_2d_np_array: N x 128 matrix of signatures
                 (where N is the number of signatures in the current session)
        :param person_sigs_np_array: M x 128 matrix of signatures
                 (where M is the number of signatures stored for current person)
        :return: score
        """
        # N x M distance matrix
        dist_matrix = self.calc_distances(
                session_sigs_2d_np_array, person_sigs_np_array)
        # Computes the 40th percentile value between flattened
        # session_sigs_2d_np_array and person_sigs_np_array
        score = np.percentile(dist_matrix, 40, interpolation='linear')
        return score


    def calc_distances(self, sigs_2d_np_array1, sigs_2d_np_array2):
        """
        Caluclates the L1 (aka 'cityblock') distance between the two given
        signature matricies
        :param sigs_2d_np_array1: N x 128 matrix of signatures
        :param sigs_2d_np_array2: M x 128 matrix of signatures
        :return: N x M distance matrix A,
                 where [Aij] = dist(array1[i], array2[j])
        """
        N = sigs_2d_np_array1.shape[0]
        M = sigs_2d_np_array2.shape[0]
        output = np.zeros((N,M))
        for m in range(M):
            for n in range(N):
                v1 = sigs_2d_np_array1[n,:]
                v2 = sigs_2d_np_array2[m,:]
                output[n,m] = np.sum(np.abs(v1-v2))

        output = output/float(self.SIGNATURE_LEN)

        # the following does the same using scipy, not used here for better compatabilty with windows
        # return (cdist(sigs_2d_np_array1, sigs_2d_np_array2, metric='cityblock')
        #         / self.SIGNATURE_LEN)

        return output


    def find_best_person_for_sigs(self):
        """
        Looks for a person from all people stored in the database that most
        closely matches the current session.
        :return: the person most closely matching faces in the session, or None
                 if no person matches within a SESSION_TO_PERSON_THRESHOLD
        """

        min_dist = 10e9
        best_person = None
        people_list = self.db_manager.get_all_people()
        for person in people_list:
            person_faces = person.faces
            person_sigs = [face.get_signature() for face in person_faces]
            if len(person_sigs) == 0:
                continue
            # Convert person_sigs to a 2d numpy array:
            person_sigs_2d_nparray = self.sig_list_to_2d_numpy_array(person_sigs)
            curr_person_dist = self.compare_session_sigs_to_person_sigs(
                        self.session_sigs_2d_nparray, person_sigs_2d_nparray)
            print 'dist of person id {} to session sigs: {}' \
                  ''.format(person.id, curr_person_dist)

            if curr_person_dist < min_dist:
                min_dist = curr_person_dist
                best_person = person

        if (min_dist <= self.SESSION_TO_PERSON_THRESHOLD):
            self.db_person_id = best_person.id
            self.db_person_prob = 100 - min_dist
            return best_person, self.db_person_prob
        else:
            self.db_person_id = None
            self.db_person_prob = 0
            return None, 0

    def get_sig_to_session_score(self, signature):
        """
        Calculates the match score between <signature> and
        signatures stored in the current session
        :param signature: Face signature (int array)
        """
        dists = self.calc_distances(np.array([signature]),
                                    self.session_sigs_2d_nparray)
        score = dists.mean()
        print 'score of sig to session: {}'.format(score)
        return score


"""
MODIFIED Bluegiga BGAPI/BGLib implementation
============================================
Bluegiga BGLib Python interface library
2013-05-04 by Jeff Rowberg <jeff@rowberg.net>
Updates should (hopefully) always be available at
https://github.com/jrowberg/bglib
Thanks to Masaaki Shibata for Python event handler code
http://www.emptypage.jp/notes/pyevent.en.html
============================================
BGLib Python interface library code is placed under the MIT license
Copyright (c) 2013 Jeff Rowberg

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
===============================================
"""

from __future__ import print_function

import logging
from struct import unpack
from enum import Enum

log = logging.getLogger(__name__)


ResponsePacketType = Enum('ResponsePacketType', [
    'system_reset',
    'system_hello',
    'system_address_get',
    'system_reg_write',
    'system_reg_read',
    'system_get_counters',
    'system_get_connections',
    'system_read_memory',
    'system_get_info',
    'system_endpoint_tx',
    'system_whitelist_append',
    'system_whitelist_remove',
    'system_whitelist_clear',
    'system_endpoint_rx',
    'system_endpoint_set_watermarks',
    'flash_ps_defrag',
    'flash_ps_dump',
    'flash_ps_erase_all',
    'flash_ps_save',
    'flash_ps_load',
    'flash_ps_erase',
    'flash_erase_page',
    'flash_write_words',
    'attributes_write',
    'attributes_read',
    'attributes_read_type',
    'attributes_user_read_response',
    'attributes_user_write_response',
    'connection_disconnect',
    'connection_get_rssi',
    'connection_update',
    'connection_version_update',
    'connection_channel_map_get',
    'connection_channel_map_set',
    'connection_features_get',
    'connection_get_status',
    'connection_raw_tx',
    'attclient_find_by_type_value',
    'attclient_read_by_group_type',
    'attclient_read_by_type',
    'attclient_find_information',
    'attclient_read_by_handle',
    'attclient_attribute_write',
    'attclient_write_command',
    'attclient_indicate_confirm',
    'attclient_read_long',
    'attclient_prepare_write',
    'attclient_execute_write',
    'attclient_read_multiple',
    'sm_encrypt_start',
    'sm_set_bondable_mode',
    'sm_delete_bonding',
    'sm_set_parameters',
    'sm_passkey_entry',
    'sm_get_bonds',
    'sm_set_oob_data',
    'gap_set_privacy_flags',
    'gap_set_mode',
    'gap_discover',
    'gap_connect_direct',
    'gap_end_procedure',
    'gap_connect_selective',
    'gap_set_filtering',
    'gap_set_scan_parameters',
    'gap_set_adv_parameters',
    'gap_set_adv_data',
    'gap_set_directed_connectable_mode',
    'hardware_io_port_config_irq',
    'hardware_set_soft_timer',
    'hardware_adc_read',
    'hardware_io_port_config_direction',
    'hardware_io_port_config_function',
    'hardware_io_port_config_pull',
    'hardware_io_port_write',
    'hardware_io_port_read',
    'hardware_spi_config',
    'hardware_spi_transfer',
    'hardware_i2c_read',
    'hardware_i2c_write',
    'hardware_set_txpower',
    'hardware_timer_comparator',
    'test_phy_tx',
    'test_phy_rx',
    'test_phy_end',
    'test_phy_reset',
    'test_get_channel_map',
    'test_debug',
    ])


EventPacketType = Enum('EventPacketType', [
    'system_boot',
    'system_debug',
    'system_endpoint_watermark_rx',
    'system_endpoint_watermark_tx',
    'system_script_failure',
    'system_no_license_key',
    'flash_ps_key',
    'attributes_value',
    'attributes_user_read_request',
    'attributes_status',
    'connection_status',
    'connection_version_ind',
    'connection_feature_ind',
    'connection_raw_rx',
    'connection_disconnected',
    'attclient_indicated',
    'attclient_procedure_completed',
    'attclient_group_found',
    'attclient_attribute_found',
    'attclient_find_information_found',
    'attclient_attribute_value',
    'attclient_read_multiple_response',
    'sm_smp_data',
    'sm_bonding_fail',
    'sm_passkey_display',
    'sm_passkey_request',
    'sm_bond_status',
    'gap_scan_response',
    'gap_mode_changed',
    'hardware_io_port_status',
    'hardware_soft_timer',
    'hardware_adc_result',
    ])

# Map a tuple of (class, command) to an enum identifier for the packet
RESPONSE_PACKET_MAPPING = {
    (0, 0): ResponsePacketType.system_reset,
    (0, 1): ResponsePacketType.system_hello,
    (0, 2): ResponsePacketType.system_address_get,
    (0, 3): ResponsePacketType.system_reg_write,
    (0, 4): ResponsePacketType.system_reg_read,
    (0, 5): ResponsePacketType.system_get_counters,
    (0, 6): ResponsePacketType.system_get_connections,
    (0, 7): ResponsePacketType.system_read_memory,
    (0, 8): ResponsePacketType.system_get_info,
    (0, 9): ResponsePacketType.system_endpoint_tx,
    (0, 10): ResponsePacketType.system_whitelist_append,
    (0, 11): ResponsePacketType.system_whitelist_remove,
    (0, 12): ResponsePacketType.system_whitelist_clear,
    (0, 13): ResponsePacketType.system_endpoint_rx,
    (0, 14): ResponsePacketType.system_endpoint_set_watermarks,

    (1, 0): ResponsePacketType.flash_ps_defrag,
    (1, 1): ResponsePacketType.flash_ps_dump,
    (1, 2): ResponsePacketType.flash_ps_erase_all,
    (1, 3): ResponsePacketType.flash_ps_save,
    (1, 4): ResponsePacketType.flash_ps_load,
    (1, 5): ResponsePacketType.flash_ps_erase,
    (1, 6): ResponsePacketType.flash_erase_page,
    (1, 7): ResponsePacketType.flash_write_words,

    (2, 0): ResponsePacketType.attributes_write,
    (2, 1): ResponsePacketType.attributes_read,
    (2, 2): ResponsePacketType.attributes_read_type,
    (2, 3): ResponsePacketType.attributes_user_read_response,
    (2, 4): ResponsePacketType.attributes_user_write_response,

    (3, 0): ResponsePacketType.connection_disconnect,
    (3, 1): ResponsePacketType.connection_get_rssi,
    (3, 2): ResponsePacketType.connection_update,
    (3, 3): ResponsePacketType.connection_version_update,
    (3, 4): ResponsePacketType.connection_channel_map_get,
    (3, 5): ResponsePacketType.connection_channel_map_set,
    (3, 6): ResponsePacketType.connection_features_get,
    (3, 7): ResponsePacketType.connection_get_status,
    (3, 8): ResponsePacketType.connection_raw_tx,

    (4, 0): ResponsePacketType.attclient_find_by_type_value,
    (4, 1): ResponsePacketType.attclient_read_by_group_type,
    (4, 2): ResponsePacketType.attclient_read_by_type,
    (4, 3): ResponsePacketType.attclient_find_information,
    (4, 4): ResponsePacketType.attclient_read_by_handle,
    (4, 5): ResponsePacketType.attclient_attribute_write,
    (4, 6): ResponsePacketType.attclient_write_command,
    (4, 7): ResponsePacketType.attclient_indicate_confirm,
    (4, 8): ResponsePacketType.attclient_read_long,
    (4, 9): ResponsePacketType.attclient_prepare_write,
    (4, 10): ResponsePacketType.attclient_execute_write,
    (4, 10): ResponsePacketType.attclient_execute_write,

    (5, 0): ResponsePacketType.sm_encrypt_start,
    (5, 1): ResponsePacketType.sm_set_bondable_mode,
    (5, 2): ResponsePacketType.sm_delete_bonding,
    (5, 3): ResponsePacketType.sm_set_parameters,
    (5, 4): ResponsePacketType.sm_passkey_entry,
    (5, 5): ResponsePacketType.sm_get_bonds,
    (5, 6): ResponsePacketType.sm_set_oob_data,

    (6, 0): ResponsePacketType.gap_set_privacy_flags,
    (6, 1): ResponsePacketType.gap_set_mode,
    (6, 2): ResponsePacketType.gap_discover,
    (6, 3): ResponsePacketType.gap_connect_direct,
    (6, 4): ResponsePacketType.gap_end_procedure,
    (6, 5): ResponsePacketType.gap_connect_selective,
    (6, 6): ResponsePacketType.gap_set_filtering,
    (6, 7): ResponsePacketType.gap_set_scan_parameters,
    (6, 8): ResponsePacketType.gap_set_adv_parameters,
    (6, 9): ResponsePacketType.gap_set_adv_data,
    (6, 10): ResponsePacketType.gap_set_directed_connectable_mode,

    (7, 0): ResponsePacketType.hardware_io_port_config_irq,
    (7, 1): ResponsePacketType.hardware_set_soft_timer,
    (7, 2): ResponsePacketType.hardware_adc_read,
    (7, 3): ResponsePacketType.hardware_io_port_config_direction,
    (7, 4): ResponsePacketType.hardware_io_port_config_function,
    (7, 5): ResponsePacketType.hardware_io_port_config_pull,
    (7, 6): ResponsePacketType.hardware_io_port_write,
    (7, 7): ResponsePacketType.hardware_io_port_read,
    (7, 8): ResponsePacketType.hardware_spi_config,
    (7, 9): ResponsePacketType.hardware_spi_transfer,
    (7, 10): ResponsePacketType.hardware_i2c_read,
    (7, 11): ResponsePacketType.hardware_i2c_write,
    (7, 12): ResponsePacketType.hardware_set_txpower,
    (7, 13): ResponsePacketType.hardware_timer_comparator,

    (8, 0): ResponsePacketType.test_phy_tx,
    (8, 1): ResponsePacketType.test_phy_rx,
    (8, 2): ResponsePacketType.test_phy_reset,
    (8, 3): ResponsePacketType.test_get_channel_map,
    (8, 4): ResponsePacketType.test_debug,
}

# TODO instead of this, have a different enum for each message type + class, and
# then just index into it

EVENT_PACKET_MAPPING = {
    (0, 0): EventPacketType.system_boot,
    (0, 1): EventPacketType.system_debug,
    (0, 2): EventPacketType.system_endpoint_watermark_rx,
    (0, 3): EventPacketType.system_endpoint_watermark_tx,
    (0, 4): EventPacketType.system_script_failure,
    (0, 5): EventPacketType.system_no_license_key,

    (1, 0): EventPacketType.flash_ps_key,

    (2, 0): EventPacketType.attributes_value,
    (2, 1): EventPacketType.attributes_user_read_request,
    (2, 2): EventPacketType.attributes_status,

    (3, 0): EventPacketType.connection_status,
    (3, 1): EventPacketType.connection_version_ind,
    (3, 2): EventPacketType.connection_feature_ind,
    (3, 3): EventPacketType.connection_raw_rx,
    (3, 4): EventPacketType.connection_disconnected,

    (4, 0): EventPacketType.attclient_indicated,
    (4, 1): EventPacketType.attclient_procedure_completed,
    (4, 2): EventPacketType.attclient_group_found,
    (4, 3): EventPacketType.attclient_attribute_found,
    (4, 4): EventPacketType.attclient_find_information_found,
    (4, 5): EventPacketType.attclient_attribute_value,
    (4, 6): EventPacketType.attclient_read_multiple_response,

    (5, 0): EventPacketType.sm_smp_data,
    (5, 1): EventPacketType.sm_bonding_fail,
    (5, 2): EventPacketType.sm_passkey_display,
    (5, 3): EventPacketType.sm_passkey_request,
    (5, 4): EventPacketType.sm_bond_status,

    (6, 0): EventPacketType.gap_scan_response,
    (6, 1): EventPacketType.gap_mode_changed,

    (7, 0): EventPacketType.hardware_io_port_status,
    (7, 1): EventPacketType.hardware_soft_timer,
    (7, 2): EventPacketType.hardware_adc_result,
}


class BGLib(object):
    """
    Modified version of jrowberg's BGLib implementation.
    """
    def __init__(self):
        self.buffer = []
        self.expected_length = 0
        # Packet message types
        self._ble_event = 0x80
        self._ble_response = 0x00
        self._wifi_event = 0x88
        self._wifi_response = 0x08

    def send_command(self, ser, packet):
        """
        Send a packet to the BLED12 over serial.

        ser -- The serial.Serial object to write to.
        packet -- The packet to write.
        """
        try:
            ser.write(packet)
        except Exception as ex:
            log.error('EXCEPTION when sending command over serial to MyMe device.'
                      '\nException Message: {}'.format(ex))
            raise ex


    def parse_byte(self, byte):
        """
        Re-build packets read in from bytes over serial one byte at a time.

        byte -- the next byte to add to the packet.

        Returns a list of the bytes in the packet once a full packet is read.
        Returns None otherwise.
        """
        if (len(self.buffer) == 0 and
            (byte == self._ble_event or byte == self._ble_response or
             byte == self._wifi_event or byte == self._wifi_response)):
            self.buffer.append(byte)
        elif len(self.buffer) == 1:
            self.buffer.append(byte)
            self.expected_length = 4 +\
                (self.buffer[0] & 0x07) + self.buffer[1]
        elif len(self.buffer) > 1:
            self.buffer.append(byte)

        if (self.expected_length > 0 and
                len(self.buffer) == self.expected_length):
            packet = self.buffer
            self.buffer = []
            return packet
        return None

    def _decode_response_packet(self, packet_class, packet_command, payload,
                                payload_length):
        packet_type = RESPONSE_PACKET_MAPPING.get(
            (packet_class, packet_command))
        if packet_type is None:
            # TODO unrecognized packet, log something?
            return

        response = {}
        if packet_type == ResponsePacketType.system_address_get:
            address = unpack('<6s', payload[:6])[0]
            address = [ord(b) for b in address]
            response = {
                'address': address
            }
        elif packet_type == ResponsePacketType.system_reg_read:
            address, value =\
                unpack('<HB', payload[:3])
            response = {
                'address': address, 'value': value
            }
        elif packet_type == ResponsePacketType.system_get_counters:
            txok, txretry, rxok, rxfail, mbuf =\
                unpack('<BBBBB', payload[:5])
            response = {
                'txok': txok, 'txretry': txretry, 'rxok': rxok,
                'rxfail': rxfail, 'mbuf': mbuf
            }
        elif packet_type == ResponsePacketType.system_get_connections:
            maxconn = unpack('<B', payload[:1])[0]
            response = {
                'maxconn': maxconn
            }
        elif packet_type == ResponsePacketType.system_read_memory:
            address, data_len =\
                unpack('<IB', payload[:5])
            data_data = [ord(b) for b in payload[5:]]
            response = {
                'address': address, 'data': data_data
            }
        elif packet_type == ResponsePacketType.system_get_info:
            data = unpack('<HHHHHBB', payload[:12])
            response = {
                'major': data[0], 'minor': data[1],
                'patch': data[2], 'build': data[3],
                'll_version': data[4], 'protocol_version': data[5],
                'hw': data[6]
            }
        elif packet_type in [
                ResponsePacketType.system_endpoint_tx,
                ResponsePacketType.system_whitelist_append,
                ResponsePacketType.system_whitelist_remove,
                ResponsePacketType.system_endpoint_set_watermarks,
                ResponsePacketType.flash_ps_save,
                ResponsePacketType.flash_erase_page,
                ResponsePacketType.attributes_write,
                ResponsePacketType.system_reg_write,
                ResponsePacketType.attclient_indicate_confirm,
                ResponsePacketType.sm_delete_bonding,
                ResponsePacketType.sm_passkey_entry,
                ResponsePacketType.gap_set_mode,
                ResponsePacketType.gap_discover,
                ResponsePacketType.gap_end_procedure,
                ResponsePacketType.gap_set_filtering,
                ResponsePacketType.hardware_timer_comparator,
                ResponsePacketType.test_phy_end,
                ResponsePacketType.hardware_spi_config,
                ResponsePacketType.gap_set_scan_parameters,
                ResponsePacketType.gap_set_adv_parameters,
                ResponsePacketType.gap_set_adv_data,
                ResponsePacketType.gap_set_directed_connectable_mode,
                ResponsePacketType.hardware_io_port_config_irq,
                ResponsePacketType.hardware_set_soft_timer,
                ResponsePacketType.hardware_adc_read,
                ResponsePacketType.hardware_io_port_config_direction,
                ResponsePacketType.hardware_io_port_config_function,
                ResponsePacketType.hardware_io_port_config_pull,
                ResponsePacketType.hardware_io_port_write]:
            result = unpack('<H', payload[:2])[0]
            response = {
                'result': result
            }
        elif packet_type == ResponsePacketType.system_endpoint_rx:
            result, data_len =\
                unpack('<HB', payload[:3])
            data_data = [ord(b) for b in payload[3:]]
            response = {
                'result': result, 'data': data_data
            }
        elif packet_type == ResponsePacketType.flash_ps_load:
            result, value_len = unpack('<HB',
                                       payload[:3])
            value_data = [ord(b) for b in payload[3:]]
            response = {
                'result': result, 'value': value_data
            }
        elif packet_type == ResponsePacketType.attributes_read:
            handle, offset, result, value_len = unpack(
                '<HHHB', payload[:7]
            )
            value_data = [ord(b) for b in payload[7:]]
            response = {
                'handle': handle, 'offset': offset,
                'result': result, 'value': value_data
            }
        elif packet_type == ResponsePacketType.attributes_read_type:
            handle, result, value_len = unpack(
                '<HHB', payload[:5]
            )
            value_data = [ord(b) for b in payload[5:]]
            response = {
                'handle': handle, 'result': result,
                'value': value_data
            }
        elif packet_type in [
                ResponsePacketType.connection_disconnect,
                ResponsePacketType.connection_update,
                ResponsePacketType.connection_version_update,
                ResponsePacketType.connection_channel_map_set,
                ResponsePacketType.connection_features_get,
                ResponsePacketType.attclient_find_by_type_value,
                ResponsePacketType.attclient_read_by_group_type,
                ResponsePacketType.attclient_read_by_type,
                ResponsePacketType.attclient_find_information,
                ResponsePacketType.attclient_read_by_handle,
                ResponsePacketType.attclient_attribute_write,
                ResponsePacketType.attclient_write_command,
                ResponsePacketType.attclient_read_long,
                ResponsePacketType.attclient_prepare_write,
                ResponsePacketType.attclient_execute_write,
                ResponsePacketType.attclient_read_multiple,
                ]:
            connection, result = unpack(
                '<BH', payload[:3]
            )
            response = {
                'connection_handle': connection, 'result': result
            }
        elif packet_type == ResponsePacketType.connection_get_rssi:
            connection, rssi = unpack(
                '<Bb', payload[:2]
            )
            response = {
                'connection_handle': connection, 'rssi': rssi
            }
        elif packet_type == ResponsePacketType.connection_channel_map_get:
            connection, map_len = unpack(
                '<BB', payload[:2]
            )
            map_data = [ord(b) for b in payload[2:]]
            response = {
                'connection_handle': connection, 'map': map_data
            }
        elif packet_type == ResponsePacketType.connection_get_status:
            connection = unpack('<B', payload[:1])[0]
            response = {
                'connection_handle': connection
            }
        elif packet_type == ResponsePacketType.connection_raw_tx:
            connection = unpack('<B', payload[:1])[0]
            response = {
                'connection_handle': connection
            }
        elif packet_type == ResponsePacketType.sm_encrypt_start:
            handle, result = unpack(
                '<BH', payload[:3]
            )
            response = {
                'handle': handle, 'result': result
            }
        elif packet_type == ResponsePacketType.sm_get_bonds:
            bonds = unpack('<B', payload[:1])[0]
            response = {
                'bonds': bonds
            }
        elif packet_type == ResponsePacketType.gap_connect_direct:
            result, connection_handle = unpack(
                '<HB', payload[:3]
            )
            response = {
                'result': result,
                'connection_handle': connection_handle
            }
        elif packet_type == ResponsePacketType.gap_connect_selective:
            result, connection_handle = unpack(
                '<HB', payload[:3]
            )
            response = {
                'result': result,
                'connection_handle': connection_handle
            }
        elif packet_type == ResponsePacketType.hardware_io_port_read:
            result, port, data = unpack(
                '<HBB', payload[:4]
            )
            response = {
                'result': result, 'port': port, 'data': data
            }
        elif packet_type == ResponsePacketType.hardware_spi_transfer:
            result, channel, data_len = unpack(
                '<HBB', payload[:4]
            )
            data_data = [ord(b) for b in payload[4:]]
            response = {
                'result': result, 'channel': channel,
                'data': data_data
            }
        elif packet_type == ResponsePacketType.hardware_i2c_read:
            result, data_len = unpack(
                '<HB', payload[:3]
            )
            data_data = [ord(b) for b in payload[3:]]
            response = {
                'result': result, 'data': data_data
            }
        elif packet_type == ResponsePacketType.hardware_i2c_write:
            written = unpack('<B', payload[:1])[0]
            response = {
                'written': written
            }
        elif packet_type == ResponsePacketType.test_get_channel_map:
            # channel_map_len = unpack(
            #    '<B', payload[:1]
            # )[0]
            channel_map_data =\
                [ord(b) for b in payload[1:]]
            response = {
                'channel_map': channel_map_data
            }
        elif packet_type == ResponsePacketType.test_debug:
            # output_len = unpack('<B',
            #                     payload[:1])[0]
            output_data =\
                [ord(b) for b in payload[1:]]
            response = {
                'output': output_data
            }

        return packet_type, response

    def _decode_event_packet(self, packet_class, packet_command, payload,
                             payload_length):
        packet_type = EVENT_PACKET_MAPPING.get((packet_class, packet_command))
        if packet_type is None:
            # TODO unrecognized packet, log something?
            return

        response = {}
        if packet_type == EventPacketType.system_boot:
            data = unpack('<HHHHHBB', payload[:12])
            response = {
                'major': data[0], 'minor': data[1],
                'patch': data[2], 'build': data[3],
                'll_version': data[4], 'protocol_version': data[5],
                'hw': data[6]
            }
        elif packet_type == EventPacketType.system_debug:
            data_len = unpack('<B', payload[:1])[0]
            data_data = [ord(b) for b in payload[1:]]
            response = {
                'data': data_data
            }
        elif packet_type in [EventPacketType.system_endpoint_watermark_rx,
                             EventPacketType.system_endpoint_watermark_tx
                             ]:
            endpoint, data = unpack(
                '<BB', payload[:2]
            )
            response = {
                'endpoint': endpoint, 'data': data
            }
        elif packet_type == EventPacketType.system_script_failure:
            address, reason = unpack(
                '<HH', payload[:4]
            )
            response = {
                'address': address, 'reason': reason
            }
        elif packet_type == EventPacketType.flash_ps_key:
            key, value_len = unpack(
                '<HB', payload[:3]
            )
            value_data = [ord(b) for b in payload[3:]]
            response = {
                'key': key, 'value': value_data
            }
        elif packet_type == EventPacketType.attributes_value:
            connection, reason, handle, offset, value_len = unpack(
                '<BBHHB', payload[:7]
            )
            value_data = [ord(b) for b in payload[7:]]
            response = {
                'connection_handle': connection, 'reason': reason,
                'handle': handle, 'offset': offset,
                'value': value_data
            }
        elif packet_type == EventPacketType.attributes_user_read_request:
            connection, handle, offset, maxsize = unpack(
                '<BHHB', payload[:6]
            )
            response = {
                'connection_handle': connection, 'handle': handle,
                'offset': offset, 'maxsize': maxsize
            }
        elif packet_type == EventPacketType.attributes_status:
            handle, flags = unpack('<HB', payload[:3])
            response = {
                'handle': handle, 'flags': flags
            }
        elif packet_type == EventPacketType.connection_status:
            data = unpack('<BB6sBHHHB', payload[:16])
            address = [ord(b) for b in data[2]]
            response = {
                'connection_handle': data[0], 'flags': data[1],
                'address': address, 'address_type': data[3],
                'conn_interval': data[4], 'timeout': data[5],
                'latency': data[6], 'bonding': data[7]
            }
        elif packet_type == EventPacketType.connection_version_ind:
            connection, vers_nr, comp_id, sub_vers_nr = unpack(
                '<BBHH', payload[:6]
            )
            response = {
                'connection_handle': connection, 'vers_nr': vers_nr,
                'comp_id': comp_id, 'sub_vers_nr': sub_vers_nr
            }
        elif packet_type == EventPacketType.connection_feature_ind:
            connection, features_len = unpack(
                '<BB', payload[:2]
            )
            features_data =\
                [ord(b) for b in payload[2:]]
            response = {
                'connection_handle': connection, 'features': features_data
            }
        elif packet_type == EventPacketType.connection_raw_rx:
            connection, data_len = unpack(
                '<BB', payload[:2]
            )
            data_data = [ord(b) for b in payload[2:]]
            response = {
                'connection_handle': connection, 'data': data_data
            }
        elif packet_type == EventPacketType.connection_disconnected:
            connection, reason = unpack(
                '<BH', payload[:3]
            )
            response = {
                'connection_handle': connection, 'reason': reason
            }
        elif packet_type == EventPacketType.attclient_indicated:
            connection, attrhandle = unpack(
                '<BH', payload[:3]
            )
            response = {
                'connection_handle': connection, 'attrhandle': attrhandle
            }
        elif packet_type == EventPacketType.attclient_procedure_completed:
            connection, result, chrhandle = unpack(
                '<BHH', payload[:5]
            )
            response = {
                'connection_handle': connection, 'result': result,
                'chrhandle': chrhandle
            }
        elif packet_type == EventPacketType.attclient_group_found:
            connection, start, end, uuid_len = unpack(
                '<BHHB', payload[:6]
            )
            uuid_data = [ord(b) for b in payload[6:]]
            response = {
                'connection_handle': connection, 'start': start,
                'end': end, 'uuid': uuid_data
            }
        elif packet_type == EventPacketType.attclient_attribute_found:
            data = unpack('<BHHBB', payload[:7])
            uuid_data = [ord(b) for b in payload[7:]]
            response = {
                'connection_handle': data[0], 'chrdecl': data[1],
                'value': data[2], 'properties': data[3],
                'uuid': uuid_data
            }
        elif packet_type == EventPacketType.attclient_find_information_found:
            connection, chrhandle, uuid_len = unpack(
                '<BHB', payload[:4]
            )
            uuid_data = [ord(b) for b in payload[4:]]
            response = {
                'connection_handle': connection, 'chrhandle': chrhandle,
                'uuid': uuid_data
            }
        elif packet_type == EventPacketType.attclient_attribute_value:
            connection, atthandle, type, value_len = unpack(
                '<BHBB', payload[:5]
            )
            value_data = [ord(b) for b in payload[5:]]
            response = {
                'connection_handle': connection, 'atthandle': atthandle,
                'type': type, 'value': value_data
            }
        elif packet_type == EventPacketType.attclient_read_multiple_response:
            connection, handles_len = unpack(
                '<BB', payload[:2]
            )
            handles_data =\
                [ord(b) for b in payload[2:]]
            response = {
                'connection_handle': connection, 'handles': handles_data
            }
        elif packet_type == EventPacketType.sm_smp_data:
            handle, packet, data_len = unpack(
                '<BBB', payload[:3]
            )
            data_data = [ord(b) for b in payload[3:]]
            response = {
                'handle': handle, 'packet': packet,
                'data': data_data
            }
        elif packet_type == EventPacketType.sm_bonding_fail:
            handle, result = unpack(
                '<BH', payload[:3]
            )
            response = {
                'handle': handle, 'result': result
            }
        elif packet_type == EventPacketType.sm_passkey_display:
            handle, passkey = unpack(
                '<BI', payload[:5]
            )
            response = {
                'handle': handle, 'passkey': passkey
            }
        elif packet_type == EventPacketType.sm_passkey_request:
            handle = unpack('<B', payload[:1])[0]
            response = {
                'handle': handle
            }
        elif packet_type == EventPacketType.sm_bond_status:
            bond, keysize, mitm, keys = unpack(
                '<BBBB', payload[:4]
            )
            response = {
                'bond': bond, 'keysize': keysize, 'mitm': mitm,
                'keys': keys
            }
        elif packet_type == EventPacketType.gap_scan_response:
            data = unpack('<bB6sBBB', payload[:11])
            sender = [ord(b) for b in data[2]]
            data_data = [ord(b) for b in payload[11:]]
            response = {
                'rssi': data[0], 'packet_type': data[1],
                'sender': sender, 'address_type': data[3],
                'bond': data[4], 'data': data_data
            }
        elif packet_type == EventPacketType.gap_mode_changed:
            discover, connect = unpack(
                '<BB', payload[:2]
            )
            response = {
                'discover': discover, 'connect': connect
            }
        elif packet_type == EventPacketType.hardware_io_port_status:
            timestamp, port, irq, state = unpack(
                '<IBBB', payload[:7]
            )
            response = {
                'timestamp': timestamp, 'port': port, 'irq': irq,
                'state': state
            }
        elif packet_type == EventPacketType.hardware_io_soft_timer:
            handle = unpack('<B', payload[:1])[0]
            response = {
                'handle': handle
            }
        elif packet_type == EventPacketType.hardware_adc_result:
            input, value = unpack('<Bh', payload[:3])
            response = {
                'input': input, 'value': value
            }

        return packet_type, response

    def decode_packet(self, packet):
        """
        Decode the packet and call the appropriate handler for the packet type.

        packet -- a list of bytes in the packet to decode.

        Returns a tuple of (PacketType, dict response data)

          BGAPI packet structure (as of 2012-11-07):
            Byte 0:
                  [7] - 1 bit, Message Type (MT)     Command/Response, 1 = Event
                [6:3] - 4 bits, Technology Type (TT)    0000 = BLE, 0001 = Wi-Fi
                [2:0] - 3 bits, Length High (LH)      Payload length (high bits)
            Byte 1:     8 bits, Length Low (LL)        Payload length (low bits)
            Byte 2:     8 bits, Class ID (CID)          Command class ID
            Byte 3:     8 bits, Command ID (CMD)         Command ID
            Bytes 4-n:  0 - 2048 Bytes, Payload (PL) Up to 2048 bytes of payload
        """
        packet_id, payload_length, packet_class, packet_command = packet[:4]
        # TODO we are not parsing out the high bits of the payload length from
        # the first byte
        payload = b''.join(chr(i) for i in packet[4:])
        message_type = packet_id & 0x88
        if message_type == 0:
            return self._decode_response_packet(
                packet_class, packet_command, payload, payload_length)
        elif message_type == 0x80:
            return self._decode_event_packet(
                packet_class, packet_command, payload, payload_length)



from __future__ import print_function

import re
import logging
import platform
import sys
import time
import threading
import subprocess
from uuid import UUID
try:
    import pexpect
except Exception as e:
    if platform.system() != 'Windows':
        print("WARNING:", e, file=sys.stderr)

from ...exceptions import (NotConnectedError, NotificationTimeout, BLEError,
                               NoResponseError)
from .. import BLEBackend, Characteristic
from ..backend import DEFAULT_CONNECT_TIMEOUT_S
from .device import GATTToolBLEDevice

log = logging.getLogger(__name__)

DEFAULT_TIMEOUT_S = 0.5


def at_most_one_device(func):
    """Every connection-specific function on the backend takes an instance of
    GATTToolBLEDevice as the first argument - this decorator will raise an
    exception if that device is not what the backend thinks is the currently
    connected device.
    """
    def wrapper(self, connected_device, *args, **kwargs):
        if connected_device != self._connected_device:
            raise NotConnectedError()
        return func(self, *args, **kwargs)
    return wrapper


class GATTToolBackend(BLEBackend):
    """
    Backend to pygatt that uses BlueZ's interactive gatttool CLI prompt.
    """
    _GATTTOOL_PROMPT = r".*> "

    def __init__(self, hci_device='hci0', gatttool_logfile=None,
                 cli_options=None):
        """
        Initialize.

        hci_device -- the hci_device to use with GATTTool.
        gatttool_logfile -- an optional filename to store raw gatttool
                input and output.
        """
        self._hci_device = hci_device
        self._cli_options = cli_options
        self._connected_device = None
        self._gatttool_logfile = gatttool_logfile
        self._receiver = None  # background notification receiving thread
        self._con = None  # gatttool interactive session

    def supports_unbonded(self):
        return False

    def start(self, reset_on_start=True):
        if self._con and self._running.is_set():
            self.stop()

        self._running = threading.Event()
        self._running.set()
        self._connection_lock = threading.RLock()

        if reset_on_start:
            # Without restarting, sometimes when trying to bond with the
            # GATTTool backend, the entire computer will lock up.
            self.reset()

        # Start gatttool interactive session for device
        gatttool_cmd = ' '.join(filter(None, [
            'gatttool',
            self._cli_options,
            '-i',
            self._hci_device,
            '-I'
        ]))
        log.debug('gatttool_cmd=%s', gatttool_cmd)
        self._con = pexpect.spawn(gatttool_cmd, logfile=self._gatttool_logfile)
        # Wait for response
        self._con.expect(r'\[LE\]>', timeout=1)

        # Start the notification receiving thread
        self._receiver = threading.Thread(target=self._receive)
        self._receiver.daemon = True
        self._receiver.start()

    def stop(self):
        """
        Disconnects any connected device, stops the backgroud receiving thread
        and closes the spawned gatttool process.
        disconnect.
        """
        self.disconnect(self._connected_device)
        if self._running.is_set():
            log.info('Stopping')
        self._running.clear()

        if self._receiver:
            self._receiver.join()
            self._receiver = None

        if self._con and self._con.isalive():
            self._con.sendline('exit')
            while True:
                if not self._con.isalive():
                    break
                time.sleep(0.1)
            self._con.close()
            self._con = None

    def scan(self, timeout=10, run_as_root=True):
        """
        By default, scanning with gatttool requires root privileges.
        If you don't want to require root, you must add a few
        'capabilities' to your system. If you have libcap installed, run this to
        enable normal users to perform LE scanning:
            setcap 'cap_net_raw,cap_net_admin+eip' `which hcitool`

        If you do use root, the hcitool subprocess becomes more difficult to
        terminate cleanly, and may leave your Bluetooth adapter in a bad state.
        """

        cmd = 'hcitool lescan'
        if run_as_root:
            cmd = 'sudo %s' % cmd

        log.info("Starting BLE scan")
        scan = pexpect.spawn(cmd)
        # "lescan" doesn't exit, so we're forcing a timeout here:
        try:
            scan.expect('foooooo', timeout=timeout)
        except pexpect.EOF:
            message = "Unexpected error when scanning"
            if "No such device" in scan.before.decode('utf-8'):
                message = "No BLE adapter found"
            log.error(message)
            raise BLEError(message)
        except pexpect.TIMEOUT:
            devices = {}
            for line in scan.before.decode('utf-8').split('\r\n'):
                match = re.match(
                    r'(([0-9A-Fa-f][0-9A-Fa-f]:?){6}) (\(?[\w]+\)?)', line)

                if match is not None:
                    address = match.group(1)
                    name = match.group(3)
                    if name == "(unknown)":
                        name = None

                    if address in devices:
                        if (devices[address]['name'] is None) and (name is not
                                                                   None):
                            log.info("Discovered name of %s as %s",
                                     address, name)
                            devices[address]['name'] = name
                    else:
                        log.info("Discovered %s (%s)", address, name)
                        devices[address] = {
                            'address': address,
                            'name': name
                        }
            log.info("Found %d BLE devices", len(devices))
            return [device for device in devices.values()]
        return []

    def connect(self, address, timeout=DEFAULT_CONNECT_TIMEOUT_S,
                address_type='public'):
        log.info('Connecting with timeout=%s', timeout)
        self._con.sendline('sec-level low')
        self._address = address
        try:
            with self._connection_lock:
                cmd = 'connect %s %s' % (self._address, address_type)
                self._con.sendline(cmd)
                self._con.expect(b'Connection successful.*\[LE\]>', timeout)
        except pexpect.TIMEOUT:
            message = ("Timed out connecting to %s after %s seconds."
                       % (self._address, timeout))
            log.error(message)
            raise NotConnectedError(message)

        self._connected_device = GATTToolBLEDevice(address, self)
        return self._connected_device

    def clear_bond(self, address=None):
        """Use the 'bluetoothctl' program to erase a stored BLE bond.
        """
        con = pexpect.spawn('sudo bluetoothctl')
        con.expect("bluetooth", timeout=1)

        log.info("Clearing bond for %s", address)
        con.sendline("remove " + address.upper())
        try:
            con.expect(["Device has been removed",
                        "# "
                        ],
                       timeout=.5)
        except pexpect.TIMEOUT:
            log.error("Unable to remove bonds for %s: %s",
                      address, con.before)
        log.info("Removed bonds for %s", address)

    @at_most_one_device
    def disconnect(self):
        with self._connection_lock:
            # TODO with gattool from bluez 5.35, gatttol consumes 100% CPU after
            # sending "disconnect". If you let the remote device do the
            # disconnect, it doesn't. Leaving it commented out for now.
            # self._con.sendline('disconnect')
            pass
        self._connected_device = None
        # TODO make call a disconnected callback on the device, so the device
        # knows if it was async disconnected?

    @at_most_one_device
    def bond(self, *args, **kwargs):
        log.info('Bonding')
        self._con.sendline('sec-level medium')
        self._con.expect(self._GATTTOOL_PROMPT, timeout=1)

    @at_most_one_device
    def discover_characteristics(self):
        characteristics = {}
        with self._connection_lock:
            self._con.sendline('characteristics')

            timeout = 6
            while True:
                try:
                    self._con.expect(
                        r"handle: 0x([a-fA-F0-9]{4}), "
                        "char properties: 0x[a-fA-F0-9]{2}, "
                        "char value handle: 0x([a-fA-F0-9]{4}), "
                        "uuid: ([0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12})\r\n",  # noqa
                        timeout=timeout)
                except pexpect.TIMEOUT:
                    break
                except pexpect.EOF:
                    break
                else:
                    try:
                        value_handle = int(self._con.match.group(2), 16)
                        char_uuid = (
                            self._con.match.group(3).strip().decode('ascii'))
                        characteristics[UUID(char_uuid)] = Characteristic(
                            char_uuid, value_handle)
                        log.debug(
                            "Found characteristic %s, value handle: 0x%x",
                            char_uuid,
                            value_handle)

                        # The characteristics all print at once, so after
                        # waiting 1-2 seconds for them to all fetch, you can
                        # load the rest without much delay at all.
                        timeout = .01
                    except AttributeError:
                        pass
        return characteristics

    def _expect(self, expected, timeout=DEFAULT_TIMEOUT_S):
        """
        We may (and often do) get an indication/notification before a
        write completes, and so it can be lost if we "expect()"'d something
        that came after it in the output, e.g.:
        > char-write-req 0x1 0x2
        Notification handle: xxx
        Write completed successfully.
        >
        Anytime we expect something we have to expect noti/indication first for
        a short time.
        """
        with self._connection_lock:
            patterns = [
                expected,
                'Notification handle = .*? \r',
                'Indication   handle = .*? \r',
                '.*Invalid file descriptor.*',
                '.*Disconnected\r',
            ]
            while True:
                try:
                    matched_pattern_index = self._con.expect(patterns, timeout)
                    if matched_pattern_index == 0:
                        break
                    elif matched_pattern_index in [1, 2]:
                        self._handle_notification_string(self._con.after)
                    elif matched_pattern_index == 3:
                        if self._running.is_set():
                            log.info("Disconnected")
                except pexpect.TIMEOUT:
                    raise NotificationTimeout(
                        "Timed out waiting for a notification")

    def _handle_notification_string(self, msg):
        hex_handle, _, hex_value = msg.strip().split()[3:]
        handle = int(hex_handle, 16)
        value = bytearray(hex_value)
        if self._connected_device is not None:
            self._connected_device.receive_notification(handle, value)

    @at_most_one_device
    def char_write_handle(self, handle, value, wait_for_response=False):
        """
        Writes a value to a given characteristic handle.
        :param handle:
        :param value:
        :param wait_for_response:
        """
        with self._connection_lock:
            hexstring = ''.join('%02x' % byte for byte in value)

            if wait_for_response:
                cmd = 'req'
            else:
                cmd = 'cmd'

            cmd = 'char-write-%s 0x%02x %s' % (cmd, handle, hexstring)

            log.debug('Sending cmd=%s', cmd)
            self._con.sendline(cmd)

            if wait_for_response:
                try:
                    self._expect('Characteristic value written successfully')
                except NoResponseError:
                    log.error("No response received", exc_info=True)
                    raise

            log.info('Sent cmd=%s', cmd)

    @at_most_one_device
    def char_read(self, uuid):
        """
        Reads a Characteristic by uuid.
        :param uuid: UUID of Characteristic to read.
        :type uuid: str
        :return: bytearray of result.
        :rtype: bytearray
        """
        with self._connection_lock:
            self._con.sendline('char-read-uuid %s' % uuid)
            self._expect('value: .*? \r')

            rval = self._con.after.split()[1:]

            return bytearray([int(x, 16) for x in rval])

    def _receive(self):
        """
        Run a background thread to listen for notifications.
        """
        log.info('Running...')
        while self._running.is_set():
            try:
                self._expect("fooooooo", timeout=.1)
            except NotificationTimeout:
                pass
            except (NotConnectedError, pexpect.EOF):
                break
            # TODO need some delay to avoid aggresively grabbing the lock,
            # blocking out the others. worst case is 1 second delay for async
            # not received as a part of another request
            time.sleep(.01)
        log.info("Listener thread finished")

    def reset(self):
        subprocess.Popen(["sudo", "systemctl", "restart", "bluetooth"]).wait()
        subprocess.Popen([
            "sudo", "hciconfig", self._hci_device, "reset"]).wait()



import threading
import time
import traceback
import Queue
from collections import defaultdict

from packet_assembler import PacketAssembler
from ..request_types import *
from ..myme_data_objects import MyMeFace
from ..myme_device import MyMeDevice, MyMeData
from ..packet_parser import PacketParser
from ..pygatt.backends import BGAPIBackend
from ..pygatt.backends.bgapi import constants
from ..pygatt.exceptions import NotConnectedError
from ..utils.sdk_utils import (print_and_log, ERROR, EXCEPTION, int_to_bytes,
                               bytearray_to_image)


class BluetoothManager(threading.Thread):

    @staticmethod
    def _default_callback(self, *args, **kargs):
        pass

    SCAN_TIMEOUT = 5
    CONNECTION_TIMEOUT = 10
    MYME_DEVICE_NAME = 'MyMe'
    def __init__(self,
                 sdk_charger_state_updated_cb,
                 sdk_device_tapped_cb,
                 sdk_battery_state_updated_cb,
                 sdk_battery_charge_level_updated_cb,
                 sdk_device_state_updated,
                 sdk_scan_complete_cb,
                 sdk_connected_to_device_cb,
                 sdk_face_signature_cb,
                 sdk_face_image_cb,
                 sdk_full_image_cb,
                 sdk_write_to_debug_cb,
                 sdk_visual_topics_cb,
                 sdk_ocr_line_cb,
                 sdk_barcode_cb):

        threading.Thread.__init__(self)
        # Program will exit if run_all_tests thread is terminated (ie: with ctrl+c)
        self.daemon = True
        self.req_queue = Queue.Queue()
        self.lock = threading.Lock()
        self._stop_event = threading.Event()

        self.sdk_charger_state_updated_cb = sdk_charger_state_updated_cb
        self.sdk_device_tapped_cb = sdk_device_tapped_cb
        self.sdk_battery_state_updated_cb = sdk_battery_state_updated_cb
        self.sdk_battery_charge_level_updated_cb = sdk_battery_charge_level_updated_cb
        self.sdk_device_state_updated = sdk_device_state_updated
        self.sdk_scan_complete_cb = sdk_scan_complete_cb
        self.sdk_connected_to_device_cb = sdk_connected_to_device_cb
        self.sdk_face_signature_cb = sdk_face_signature_cb
        self.sdk_face_image_cb = sdk_face_image_cb
        self.sdk_full_image_cb = sdk_full_image_cb
        self.sdk_write_to_debug_cb = sdk_write_to_debug_cb
        self.sdk_visual_topics_cb = sdk_visual_topics_cb
        self.sdk_ocr_line_cb = sdk_ocr_line_cb
        self.sdk_barcode_cb = sdk_barcode_cb

        while True:
            try:
                self.ble_adapter = BGAPIBackend()
                self.ble_adapter.start()
                break
            except Exception as ex:
                print_and_log('Exception when starting BLE BGAPI Backend.\n'
                              'Exception Message: {}'.format(ex))
                print_and_log('Try restarting the MyMe device and reinserting the '
                              'BLE dongle. Will retry in 5 seconds.')
                time.sleep(5)

        self.myme_device = MyMeDevice()
        self.myme_device.device_state = DeviceState.NotFound
        self.pending_requests = defaultdict(lambda : defaultdict(list))
        # Must update when connecting to device

        self.face_signature_assembler = PacketAssembler(self.face_signature_assembler_cb)
        self.face_image_assembler = PacketAssembler(self.face_image_assembler_cb)
        self.full_image_assembler = PacketAssembler(self.full_image_assembler_cb)
        self.imagenet_assembler = PacketAssembler(self.imagenet_assembler_cb)
        self.ocr_assembler = PacketAssembler(self.ocr_assembler_cb)
        self.barcode_assembler = PacketAssembler(self.barcode_assembler_cb)


    # TODO - add_and_commit stop ability
    # TODO reconnect upon disconnection


    def run(self):
        while not self._stop_event.is_set():
            try:
                item = self.req_queue.get(2)
                with self.lock:
                    self._handle_item(item)
                self.req_queue.task_done()
            except Queue.Empty: # timeout
                continue

        # upon exit, disconnect from device if we are connected
        if self.myme_device.device_state is DeviceState.Connected:
            self.disconnect_from_device()

        self.ble_adapter.stop()

    def stop(self):
        """
        Disconnects from device (if connected) and terminates the thread by exiting
        :return:
        """
        self._stop_event.set()
        self.join()

    def get_characterstics_list(self):
        if self.myme_device.device_state is not DeviceState.Connected:
            error = 'NOTE: Requested to get characterstic list while not connected.'
            print_and_log(error)
            return

        char_list = self.myme_device.handler.discover_characteristics()
        for char_uuid_str, char_obj in char_list.iteritems():
            print_and_log("Characteristic 0x{} is handle 0x{}"
                          "".format(char_uuid_str, char_obj.handle))

            for desc_uuid_str, desc_handle in char_obj.descriptors.iteritems():
                print_and_log("Characteristic descriptor 0x{} is handle 0x{}"
                              "".format(desc_uuid_str, desc_handle))

    def get_info_from_device(self, data_member, callback):
        """
        General function which returns data from a connected MyMeDevice
        :param name of data_member to get from device
        :return: On success returns requested data
        """
        if (self.myme_device.device_state not in
                [DeviceState.Connected, DeviceState.Subscribing]):
            callback(error='ERROR: Cannot get {} - not connected to device.'.format(data_member))
            return

        if data_member == MyMeData.sw_version:
            return self.myme_device.sw_version
        elif data_member == MyMeData.hw_version:
            return self.myme_device.hw_version
        elif data_member == MyMeData.battery_charge_level:
            self.get_battery_charge_level(callback)
        elif data_member == MyMeData.battery_state:
            self.get_battery_state(callback)
        elif data_member == MyMeData.charger_state:
            self.get_charger_state(callback)
        elif data_member == MyMeData.device_time:
            self.get_device_time(callback)

    def get_device_time(self, time_callback):
        with self.lock:
            device_time = self.myme_device.handler.char_read(TIME_SERVICE_UUID)
            # TODO - parse return value
            self.myme_device.device_time = device_time
            time_callback(device_time)

    def get_battery_charge_level(self, battery_charge_level_cb):
        """
        Gets the percent charge [0, 1] of the battery from the device
        Device returns battery level in the following format:
            1 byte - battery level from 0 to 100 percent (unsigned 8 bit)
        :param battery_charge_level_cb: callback function
        """
        with self.lock:
            battery_charge_level = self.myme_device.handler.char_read(
                    BATTERY_LEVEL_CHARACTERISTIC_UUID)
            # TODO - parse return value
            self.myme_device.battery_charget_level = battery_charge_level
            if battery_charge_level_cb is not None:
                battery_charge_level_cb(battery_charge_level)

    def get_battery_state(self, battery_state_cb):
        """
        Get the BatteryState of the device battery. The possible values are:
                BatteryOK = 0
                BatteryCharging = 1
                BatteryLow = 2
                BatteryCritical = 3 # Device will shutdown upon reaching state
        :param battery_state_cb: callback function
        """
        with self.lock: # note this is the caller context -  make sure the run_all_tests thread is not accessing now the bgapi
            battery_state = self.myme_device.handler.char_read(LOW_BATTERY_CHARACTERISTIC_UUID)
            # TODO - parse return value
            self.myme_device.battery_state = battery_state
            battery_state_cb(battery_state)


    def get_charger_state(self, charger_state_cb):
        with self.lock:
            charger_status = self.myme_device.handler.char_read(
                    CHARGER_CHARACTERISTIC_UUID)
            # TODO - parse return value
            self.myme_device.charger_state = charger_status
            charger_state_cb(charger_status)

    def get_device_state(self):
        return self.myme_device.device_state

    def set_device_state(self, new_state):
        self.last_device_state = self.myme_device.device_state
        self.myme_device.device_state = new_state
        self.sdk_device_state_updated(new_state)

    def get_device_hardware_version(self):
        return self.myme_device.hw_version

    def get_device_software_version(self):
        return self.myme_device.sw_version


    def connect_to_paired_device_with_timeout(self, timeout_secs, device_addr,
                                              callback=None, block=False, tries=1):
        """

        :param timeout_secs:
        :param device_addr:
        :param callback:
        :param block: if True, function will block until the connection request
                      is handled by the request handling Thread
        :param tries: Number of times to try connecting in case of failure
                     (will only retry if <block> is True) If tries is < 1,
                     will retry infinitely
        :return:
        """
        if (timeout_secs <= 0):
            timeout_secs = self.CONNECTION_TIMEOUT

        connect_req = ConnectionRequest(timeout_secs, device_addr, callback)
        connect_req.event = threading.Event()
        self.req_queue.put(connect_req)

        if block:
            if tries < 1:
                tries = int(1e18)
            while tries:
                connect_req.event.wait()
                if connect_req.success == True:
                    return
                tries -= 1


    def disconnect_from_device(self, callback=None):
        if self.myme_device.device_state is not DeviceState.Connected:
            error = 'NOTE: Requested to disconnect while not connected.'
            print_and_log(error)
            if callback:
                callback(False, error=error)
            return

        self.myme_device.handler.disconnect()
        self.set_device_state(DeviceState.Paired)
        if callback:
            callback(True)

    def scan_for_peripherals_with_timeout(self, timeout, block= False, scan_cb=None):
        """
        """
        if (timeout <= 0):
            timeout = self.SCAN_TIMEOUT

        scan_req = ScanRequest(timeout, scan_cb)
        scan_req.event = threading.Event()
        self.req_queue.put(scan_req)
        if block:
            scan_req.event.wait()

    def write_to_debug(self, text, debug_cb=None):
        """
        Outputs <text> to the device stdin
        :param text: text to write
        :param debug_cb: optional user callback which reports whether the request
                        to write to MyMe was successful
        """
        self.req_queue.put(WriteToDebugRequest(text, debug_cb))


    def request_face_image(self, db_face_id, myme_face_id, face_image_cb=None):
        """
        Sends a request to a MyMe device to fetch a face image for <db_face_id>
        :param db_face_id: Id of the face in the Faces DB
        :param myme_face_id: MyMe application internal id of the face for which
                             to fetch an image
        :param face_image_cb: callback function
        """
        self.req_queue.put(FaceImageRequest(db_face_id=db_face_id,
                                            myme_face_id=myme_face_id,
                                            callback=face_image_cb))

    def request_full_image(self, frame_idx, full_image_cb=None):
        """
        Sends a request to a MyMe device to fetch a full image for <frame_idx>
        :param frame_idx: Index of frame for which user wants to fetch an image
        :param full_image_cb: callback function to be called after full image
                              request is handled
        """
        self.req_queue.put(FullImageRequest(frame_idx, full_image_cb))

    # ************************** PRIVATE FUNCTIONS *****************************

    def _subscribe_to_all_services(self):
        """
        Subsribes to all avilable services on a MyMe device
        :param callback_dict: dict mapping ServiceName to callback
        function for that service. For each service that has a matching callback
        function, that function will be called on any notification from the service
        """
        self.myme_device.handler.subscribe(FACE_IMAGE_CHARACTERISTIC_UUID,
                                    self.FaceImageCallback)
        self.myme_device.handler.subscribe(FACE_SIGNATURE_CHARACTERISTIC_UUID,
                                    self.FaceSignatureCallback, indication=True)
        self.myme_device.handler.subscribe(IMAGENET_CHARACTERISTIC_UUID,
                                           self.ImagenetCallback, indication=True)
        self.myme_device.handler.subscribe(JPEG_CHARACTERISTIC_UUID,
                                    self.FullImageCallback)
        self.myme_device.handler.subscribe(OCR_CHARACTERISTIC_UUID,
                                    self.OcrCallback, indication=True)
        self.myme_device.handler.subscribe(BARCODE_CHARACTERISTIC_UUID,
                                    self.BarcodeCallback, indication=True)

        self.myme_device.handler.subscribe(BATTERY_LEVEL_CHARACTERISTIC_UUID,
                                           self.battery_level_cb)
        self.myme_device.handler.subscribe(CHARGER_CHARACTERISTIC_UUID,
                                           self.charger_cb, indication=True)
        self.myme_device.handler.subscribe(TAP_CHARACTERISTIC_UUID,
                                           self.tap_cb, indication=True)
        self.myme_device.handler.subscribe(LOW_BATTERY_CHARACTERISTIC_UUID,
                                           self.low_battery_cb)


    def _subscribe_to_service(self, service_uuid, callback=None):
        """
        :param service_uuid: uuid of service to subscribe to
        :param callback: optional callback function for device to call upon
                         notification from service
        """
        if (self.myme_device.device_state not in
                [DeviceState.Connected]):
            if callback:
                error=('ERROR: Cannot subscribe to services - device'
                       'is in an incorrect state ({})'
                       ''.format(self.myme_device.device_state))
                callback(error=error)
            return

        with self.lock:
            self.set_device_state(DeviceState.Subscribing)
            self.myme_device.handler.subscribe(service_uuid, callback)
            self.set_device_state(self.myme_device.last_device_state)


    # **************************************************************************
    # *********************** USER REQUEST HANDLERS ****************************
    # **************************************************************************


    def _handle_scan_request(self, req):
        device_list = []
        error = None
        try:
            device_list = self.ble_adapter.filtered_scan(
                    self.MYME_DEVICE_NAME, req.request_timeout)
            scan_state = ScanState.SuccessDueToTimeout
        except Exception as ex:
            message = ('EXCEPTION when running _handle_scan_request\n'
                       'Exception Message: {}'.format(ex))
            traceback.print_exc()
            print_and_log(message, EXCEPTION)
            error =  message
            scan_state = ScanState.FailedWithError

        # sort by descending RSSI order
        for device in device_list:
            device['rssi'] = int(device['rssi'])

        sorted_list = sorted(device_list, key=lambda k: k['rssi'], reverse=True)

        if req.callback:
            self.sdk_scan_complete_cb(sorted_list, scan_state, error, req.callback)

        # done, signal the caller in case it is blocked
        req.event.set()

    def _handle_connection_request(self, req):
        if not self.myme_device.device_state == DeviceState.Paired:
            error=('Cannot connect to device because device is in an incorrect '
                   'state: {}. (Must be in Paired to connect'
                   ''.format(self.myme_device.device_state))
            self.sdk_connected_to_device_cb(None, error, req.callback)
            return

        self.set_device_state(DeviceState.Connecting)
        while not req.success:
            try:
                self.myme_device.handler = self.ble_adapter.connect(address=req.device_addr,
                        addr_type=constants.ble_address_type['gap_address_type_random'],
                        timeout=req.request_timeout)
                self._subscribe_to_all_services()
                self.set_device_state(DeviceState.Connected)
                self.sdk_connected_to_device_cb(req.device_addr, None, req.callback)

                self.get_characterstics_list()
                self.myme_device.fill_device_info()

                req.success = True
            except Exception as ex:
                self.set_device_state(DeviceState.Paired)
                req.success = False
                error= ('Failure when trying to connect to device at address: '
                        '{}\nException Message: {}\nRETRYING'
                        ''.format(req.device_addr, ex))
                traceback.print_exc()
                self.sdk_connected_to_device_cb(None, error, req.callback)


                time.sleep(1)
                self.ble_adapter.disconnect_all()
                self.ble_adapter.reset()

        req.event.set()

    def _handle_write_to_debug_request(self, req):
        if (self.myme_device.device_state not in
                [DeviceState.Connected]):
            print_and_log('ERROR: cannot carry out write to debug request: '
                               'Not connected to device.', ERROR)
            return

        hex_packet = bytearray(req.text)
        try:
            self.myme_device.handler.char_write(DEBUG_CHARACTERISTIC_UUID,
                                                hex_packet)
            self.sdk_write_to_debug_cb(True, req.cb)
        except Exception as ex:
            print_and_log('ERROR: Failed trying to write to debug.\n'
                          'Exception Message: {}'.format(ex), ERROR)
            self.sdk_write_to_debug_cb(False, req.cb)

    def _handle_get_face_image_request(self, req):
        if (self.myme_device.device_state not in
                [DeviceState.Connected]):
            print_and_log('ERROR: cannot carry out request face image request: '
                               'Not connected to device.', ERROR)
            return
        try:
            header = bytearray([0x00])
            face_id = bytearray(int_to_bytes(req.myme_face_id, length=4))
            timestamp = bytearray([0x00 for i in range(4)])
            hex_packet = header + face_id + timestamp
            self.pending_requests[RequestType.FaceImageRequest][req.myme_face_id].append(req)
            self.myme_device.handler.char_write(FACE_IMAGE_CHARACTERISTIC_UUID,
                                                hex_packet)
        except NotConnectedError:
            print_and_log('ERROR: tried writing to face_image characteristic '
                               'while not connected.', ERROR)


    def _handle_get_face_full_image_request(self, req):
        if (self.myme_device.device_state not in
                [DeviceState.Connected]):
             print_and_log('ERROR: cannot carry out request face image request: '
                               'Not connected to device.', ERROR)
        try:
            header = bytearray([0x00])
            frame_idx = bytearray(int_to_bytes(req.frame_idx, length=4))
            timestamp = bytearray([0x00 for i in range(4)])
            hex_packet = header + frame_idx + timestamp
            self.pending_requests[RequestType.FullImageRequest][req.frame_idx].append(req)
            self.myme_device.handler.char_write(JPEG_CHARACTERISTIC_UUID,
                                                hex_packet)
        except NotConnectedError:
            print_and_log('ERROR: tried writing to face_image characteristic '
                               'while not connected.', ERROR)

    # **************************************************************************
    # ********************* ASSEMBLER RESPONSE HANDLERS ************************
    # **************************************************************************

    def _handle_face_signature_assembler_response(self, res):
        face_id, timestamp, signature_int_array, flag = \
                       PacketParser.parse_face_signature_packet(res.packet)
        self.sdk_face_signature_cb(face_id, timestamp, signature_int_array, flag)

    def _handle_face_image_assembler_response(self, res):
        myme_face_id, detected_time, raw_image_bytearray = \
            PacketParser.parse_face_image_packet(res.packet)
        if myme_face_id is None:
            print_and_log('Error fetching face image!')
            return

        curr_req_list = \
            self.pending_requests[RequestType.FaceImageRequest].get(myme_face_id)
        for req in curr_req_list:
            # Use this opportunity to store face image into the DB
            self.sdk_face_image_cb(req.db_face_id, detected_time,
                                   raw_image_bytearray, req.callback)


    def _handle_full_image_assembler_response(self, res):
        myme_image = PacketParser.parse_full_image_packet(res.packet)
        curr_req_list = \
            self.pending_requests[RequestType.FullImageRequest].get(myme_image.frame_idx)
        for req in curr_req_list:
            self.sdk_full_image_cb(myme_image, res.callback)

    def _handle_imagenet_assembler_response(self, res):
        visual_topics_list, timestamp = \
            PacketParser.parse_imagenet_packet(res.packet)
        self.sdk_visual_topics_cb(visual_topics_list, timestamp)

    def _handle_ocr_assembler_response(self, res):
        ocr_lines_list, timestamp, is_frame_end = \
            PacketParser.parse_ocr_packet(res.packet)
        for line in ocr_lines_list:
            self.sdk_ocr_line_cb(line, timestamp)
        # if is_frame_end is true, call the callback with ocr_line = None
        if is_frame_end:
            self.sdk_ocr_line_cb(None, timestamp, is_frame_end)

    def _handle_item(self, req):
        if req.request_type == RequestType.ScanRequest:
            self._handle_scan_request(req)
        elif req.request_type == RequestType.ConnectionRequest:
            self._handle_connection_request(req)
        elif req.request_type == RequestType.WriteToDebugRequest:
            self._handle_write_to_debug_request(req)
        elif req.request_type == RequestType.FaceImageRequest:
            self._handle_get_face_image_request(req)
        elif req.request_type == RequestType.FullImageRequest:
            self._handle_get_face_full_image_request(req)
        elif req.request_type == RequestType.FaceSignatureAssemblerResponse:
            self._handle_face_signature_assembler_response(req)
        elif req.request_type == RequestType.FaceImageAssemblerResponse:
            self._handle_face_image_assembler_response(req)
        elif req.request_type == RequestType.FullImageAssemblerResponse:
            self._handle_full_image_assembler_response(req)
        elif req.request_type == RequestType.ImagenetAssemblerResponse:
            self._handle_imagenet_assembler_response(req)
        elif req.request_type == RequestType.OcrAssemblerResponse:
            self._handle_ocr_assembler_response(req)
        else:
            print_and_log('ERROR: got unsupported request of type - {}. '
                          'Ignoring.'.format(req.request_type),ERROR)


    # ***************** BGAPIE RECEIVER THREAD CALLBACKS ***********************
    # Set of callbacks for bgapi receiver thread to call upon receiving a
    # notification from a characteristic
    # **************************************************************************

    def FaceSignatureCallback(self, handle, packet):
        self.face_signature_assembler.send(packet)

    def FaceImageCallback(self, handle, packet):
        self.face_image_assembler.send(packet)

    def ImagenetCallback(self, handle, packet):
        self.imagenet_assembler.send(packet)

    def FullImageCallback(self, handle, packet):
        self.full_image_assembler.send(packet)

    def OcrCallback(self, handle, packet):
        self.ocr_assembler.send(packet)

    def BarcodeCallback(self, handle, packet):
        self.barcode_assembler.send(packet)

    # ************************** ASSEMBLER CALLBACKS ***************************
    # Set of default callbacks for the assembler to call upon receiving a complete packet
    # Each request may override the default callback with one specific to the request
    # **************************************************************************

    def face_signature_assembler_cb(self, face_sig_packet):
        face_assembler_res = FaceSignatureAssemblerResponse(face_sig_packet)
        self.req_queue.put(face_assembler_res)

    def face_image_assembler_cb(self, face_image_packet):
        face_image_res = FaceImageAssemblerResponse(face_image_packet)
        self.req_queue.put(face_image_res)

    def full_image_assembler_cb(self, full_image_packet):
        full_image_res = FullImageAssemblerResponse(full_image_packet)
        self.req_queue.put(full_image_res)

    def imagenet_assembler_cb(self, visual_topics_packet):
        imagenet_assembler_res = ImagenetAssemblerResponse(visual_topics_packet)
        self.req_queue.put(imagenet_assembler_res)

    def ocr_assembler_cb(self, packet):
        ocr_assembler_res = OcrAssemblerResponse(packet)
        self.req_queue.put(ocr_assembler_res)

    def charger_cb(self, handle, packet):
        new_charger_state = PacketParser.parse_charger_packet(packet)
        if (new_charger_state is not None):
            self.sdk_charger_state_updated_cb(new_charger_state)

    def tap_cb(self, handle, packet):
        self.sdk_device_tapped_cb()

    def low_battery_cb(self, handle, packet):
        new_battery_state = PacketParser.parse_battery_state_packet(packet)
        if new_battery_state is not None:
            self.sdk_battery_state_updated_cb(new_battery_state)

    def battery_level_cb(self, handle, packet):
        battery_level = PacketParser.parse_battery_charge_level_packet(packet)
        if battery_level is not None:
            self.sdk_battery_charge_level_updated_cb(battery_level)

    def barcode_assembler_cb(self, packet):
        (barcode, timestamp) = PacketParser.parse_barcode_packet(packet)
        self.sdk_barcode_cb(barcode, timestamp)







from enum import Enum

SDK_APPNAME = 'myme_sdk'
SDK_VERSION = '0.1'
MYME_DB_NAME = 'myme_sdk.db'


class DeviceState(Enum):
    NotFound = 1
    Paired = 2
    Connecting = 3
    Connected = 4
    # Device is about to turn off (e.g battery is close to 0%) or user requested shut down
    Shuttingdown = 5

class ChargerState(Enum):
    ChargerDisconnected = 0
    ChargerConnected = 1
    ChargerError = 2
    ChargingDone = 3

class BatteryState(Enum):
    BatteryFull = 0
    BatteryOK = 1
    BatteryLow = 2
    BatteryCritical = 3 # Device about to shutdown

# State of scan for MyMe device
class ScanState(Enum):
    FailedWithError = 0
    CanceledByUser = 1 # NOT SUPPORTED
    SuccessDueToTimeout = 2

# State of connection attempt to MyMe device
class ConnectingState(Enum):
    Success = 0
    FailedWithError = 1
    FailedDueToTimeout = 2
    CanceledByUser = 3 # NOT SUPPORTED


class ServiceUUID(Enum):
    FacesServiceUuid               		= "FB694B90-F49E-4597-8306-171BBA78F846"
    BatteryServiceUuid					= "0000180F-0000-1000-8000-00805F9B34FB"
    HardwareSoftwareServiceUuid			= "0000180A-0000-1000-8000-00805F9B34FB" #
    WordCloudServiceUuid				= "A8291D00-83C9-11E5-B96C-0002A5D5C51B"
    DebugServiceUuid					= "62967EA0-4F0C-11E5-AC7D-0002A5D5C51B"
    ChargerServiceUuid					= "FEC297BC-246B-4BF2-95BD-48353A86724D"
    TimeServiceUuid						= "C18094F9-2115-4371-AA97-123B881EE420"
    LowBatteryServiceUuid				= "37712D0D-ED77-4705-AFF3-B336F9EB87B9"
    GeneralInformationServiceUuid		= "2266F960-8847-4380-8566-7A3E7827FB91"
    TapServiceUuid						= "8D06CB14-BC4A-47EE-907F-2D3E814AEFF9"
    JpegServiceId                       = "152E2EC7-D5B9-4C96-B193-8760FA6B176E"



class MultiPacketNotifyCharactersticUUID(Enum):
    FaceImageCharacteristicUuid				= "EB6727C4-F184-497A-A656-76B0CDAC633A"
    FaceSignatureCharacteristicUuid         = "18F58CA0-44D0-11E5-9A12-0002A5D5C51B"
    ImangenetCharacteristicUuid             = "C58597C0-83C9-11E5-A0Cf-0002A5D5C51B"
    JpegCharacteristicUuid					= "8953B547-7AEC-4FA8-AC20-9F710D3FD856"
    OcrCharacteristicUuid					= "56BB0234-96E9-45DC-AC99-B2802AFF6289"


FACES_SERVICE_UUID                    = "FB694B90-F49E-4597-8306-171BBA78F846"
FACE_SIGNATURE_CHARACTERISTIC_UUID    = "18F58CA0-44D0-11E5-9A12-0002A5D5C51B"
FACE_IMAGE_CHARACTERISTIC_UUID        = "EB6727C4-F184-497A-A656-76B0CDAC633A"


HARDWARE_SOFTWARE_SERVICE_UUID        = "0000180A-0000-1000-8000-00805F9B34FB" #"180A"
BATTERY_LEVEL_SERVICE_UUID            = "0000180F-0000-1000-8000-00805F9B34FB" # "180F"
BATTERY_LEVEL_CHARACTERISTIC_UUID     = "00002A19-0000-1000-8000-00805F9B34FB" # "2A19"
HARDWARE_CHARACTERISTIC_UUID          = "00002A27-0000-1000-8000-00805F9B34FB" #"2A27"
SOFTWARE_CHARACTERISTIC_UUID          = "00002A28-0000-1000-8000-00805F9B34FB" #"2A28"


IMAGENET_SERVICE_UUID                 = "A8291D00-83C9-11E5-B96C-0002A5D5C51B"
IMAGENET_CHARACTERISTIC_UUID          = "C58597C0-83C9-11E5-A0Cf-0002A5D5C51B"

DEBUG_SERVICE_UUID                    = "62967EA0-4F0C-11E5-AC7D-0002A5D5C51B"
DEBUG_CHARACTERISTIC_UUID             = "E8775746-7836-4D91-8CAB-49FDAD594570"

CHARGER_SERVICE_UUID                  = "FEC297BC-246B-4BF2-95BD-48353A86724D"
CHARGER_CHARACTERISTIC_UUID           = "F2208A8F-A3C3-4B27-BF8F-513CC9FD6365"

TIME_SERVICE_UUID                     = "C18094F9-2115-4371-AA97-123B881EE420"
TIME_CHARACTERISTIC_UUID              = "E810CEA7-5FC3-4E3E-955D-00506117B4A2"

LOW_BATTERY_SERVICE_UUID              = "37712D0D-ED77-4705-AFF3-B336F9EB87B9"
LOW_BATTERY_CHARACTERISTIC_UUID       = "bb2ea6a9-a071-4253-b8b3-ec4c6bd4408d"

GENERAL_INFORMATION_SERVICE_UUID      = "2266F960-8847-4380-8566-7A3E7827FB91"
GENERAL_DEVICE_ID_CHARACTERISTIC_UUID = "5F34A16C-02D6-451A-BDAB-4330AFBA0264"

TAP_SERVICE_UUID                      = "8D06CB14-BC4A-47EE-907F-2D3E814AEFF9"
TAP_CHARACTERISTIC_UUID               = "E5FE7BBD-2C00-402A-B139-2769909BDA3A"

# NEW SERVICE! (Exclusive to Python SDK as of Now!!)
JPEG_SERVICE_UUID                     = "152E2EC7-D5B9-4C96-B193-8760FA6B176E"
JPEG_CHARACTERISTIC_UUID              = "8953B547-7AEC-4FA8-AC20-9F710D3FD856"

OCR_SERVICE_UUID                      = "CA644A25-FE7A-4329-90E3-B048AA5B11EE"
OCR_CHARACTERISTIC_UUID               = "56BB0234-96E9-45DC-AC99-B2802AFF6289"

BARCODE_SERVICE_UUID                  = "0D5C722F-8517-472C-B681-F67A36E36F27"
BARCODE_CHARACTERISTIC_UUID           = "99206FA0-0CAE-45D9-BBAD-70643FDBB77B"



import abc

from enum import Enum

from constants import *

class RequestType(Enum):
    ScanRequest = 0
    ConnectionRequest = 1
    WriteToDebugRequest = 2
    FaceImageRequest = 3
    FullImageRequest = 4

    FaceSignatureAssemblerResponse = 5
    FaceImageAssemblerResponse = 6
    FullImageAssemblerResponse = 7
    ImagenetAssemblerResponse = 8
    OcrAssemblerResponse = 9


class Request(object):
    __metaclass__ = abc.ABCMeta
    def __init__(self):
        self.request_type = None

    def default_cb(self, error, *args, **kargs):
        print error

class ScanRequest(Request):
    def __init__(self, timeout_sec, callback=None):
        Request.__init__(self)
        self.callback = callback
        self.request_type =  RequestType.ScanRequest
        self.request_timeout = timeout_sec
        self.event = None


class ConnectionRequest(Request):
    def __init__(self, timeout_sec, device_addr, callback):
        Request.__init__(self)
        self.callback = callback
        self.device_addr = device_addr
        self.request_type = RequestType.ConnectionRequest
        self.request_timeout = timeout_sec
        self.event = None
        self.success = None


class WriteToDebugRequest(Request):
    def __init__(self, text, callback):
        Request.__init__(self)
        self.callback = callback
        self.text = text
        self.request_type = RequestType.WriteToDebugRequest

class FaceImageRequest(Request):
    def __init__(self, db_face_id, myme_face_id, callback):
        Request.__init__(self)
        self.callback = callback
        self.db_face_id = db_face_id
        self.myme_face_id = myme_face_id
        self.request_type = RequestType.FaceImageRequest

class FullImageRequest(Request):
    def __init__(self, frame_idx, callback):
        Request.__init__(self)
        self.callback = callback
        self.frame_idx = frame_idx
        self.request_type = RequestType.FullImageRequest


# ASSEMBLER RESPONSES

class FaceSignatureAssemblerResponse(Request):
    def __init__(self, face_sig_packet):
        Request.__init__(self)
        self.packet = face_sig_packet
        self.request_type = RequestType.FaceSignatureAssemblerResponse

class FaceImageAssemblerResponse(Request):
    def __init__(self, face_img_packet, callback=None):
        Request.__init__(self)
        self.callback = callback
        self.packet = face_img_packet
        self.request_type = RequestType.FaceImageAssemblerResponse

class FullImageAssemblerResponse(Request):
    def __init__(self, full_img_packet, callback=None):
        Request.__init__(self)
        self.callback = callback
        self.packet = full_img_packet
        self.request_type = RequestType.FullImageAssemblerResponse

class ImagenetAssemblerResponse(Request):
    def __init__(self, imagenet_packet):
        Request.__init__(self)
        self.packet = imagenet_packet
        self.request_type = RequestType.ImagenetAssemblerResponse

class OcrAssemblerResponse(Request):
    def __init__(self, ocr_packet):
        Request.__init__(self)
        self.packet = ocr_packet
        self.request_type = RequestType.OcrAssemblerResponse


import logging
import time

from ... import BLEDevice, exceptions
from . import constants
from .bgapi import BGAPIError
from .error_codes import ErrorCode
from .packets import BGAPICommandPacketBuilder as CommandBuilder
from .bglib import EventPacketType, ResponsePacketType

log = logging.getLogger(__name__)


def connection_required(func):
    """Raise an exception if the device is not connected before calling the
    actual function.
    """
    def wrapper(self, *args, **kwargs):
        if self._handle is None:
            raise exceptions.NotConnectedError()
        return func(self, *args, **kwargs)
    return wrapper


class BGAPIBLEDevice(BLEDevice):
    def __init__(self, address, handle, backend):
        super(BGAPIBLEDevice, self).__init__(address)
        self._handle = handle
        self._backend = backend
        self._characteristics = {}

    @connection_required
    def bond(self, permanent=False):
        """
        Create a bond and encrypted connection with the device.
        """

        # Set to bondable mode so bonds are store permanently
        if permanent:
            self._backend.set_bondable(True)
        log.debug("Bonding to %s", self._address)
        self._backend.send_command(
            CommandBuilder.sm_encrypt_start(
                self._handle, constants.bonding['create_bonding']))
        self._backend.expect(ResponsePacketType.sm_encrypt_start)

        packet_type, response = self._backend.expect_any(
            [EventPacketType.connection_status,
             EventPacketType.sm_bonding_fail])
        if packet_type == EventPacketType.sm_bonding_fail:
            raise BGAPIError("Bonding failed")
        log.info("Bonded to %s", self._address)

    @connection_required
    def get_rssi(self):
        """
        Get the receiver signal strength indicator (RSSI) value from the device.

        Returns the RSSI as in integer in dBm.
        """
        # The BGAPI has some strange behavior where it will return 25 for
        # the RSSI value sometimes... Try a maximum of 3 times.
        for i in range(0, 3):
            self._backend.send_command(
                CommandBuilder.connection_get_rssi(self._handle))
            _, response = self._backend.expect(
                ResponsePacketType.connection_get_rssi)
            rssi = response['rssi']
            if rssi != 25:
                return rssi
            time.sleep(0.1)
        raise BGAPIError("get rssi failed")

    @connection_required
    def char_read(self, uuid):
        handle = self.get_handle(uuid)
        log.info("Reading characteristic at handle %d", handle)
        self._backend.send_command(
            CommandBuilder.attclient_read_by_handle(
                self._handle, handle))

        self._backend.expect(ResponsePacketType.attclient_read_by_handle)
        matched_packet_type, response = self._backend.expect_any(
            [EventPacketType.attclient_attribute_value,
             EventPacketType.attclient_procedure_completed])
        # TODO why not just expect *only* the attribute value response, then it
        # would time out and raise an exception if allwe got was the 'procedure
        # completed' response?
        if matched_packet_type != EventPacketType.attclient_attribute_value:
            raise BGAPIError("Unable to read characteristic")
        return bytearray(response['value'])

    @connection_required
    def char_write_handle(self, char_handle, value, wait_for_response=False):
        if wait_for_response:
            raise NotImplementedError()

        while True:
            value_list = [b for b in value]
            self._backend.send_command(
                CommandBuilder.attclient_attribute_write(
                    self._handle, char_handle, value_list))

            self._backend.expect(ResponsePacketType.attclient_attribute_write)
            packet_type, response = self._backend.expect(
                EventPacketType.attclient_procedure_completed)
            if (response['result'] !=
                    ErrorCode.insufficient_authentication.value):
                # Continue to retry until we are bonded
                break

    @connection_required
    def disconnect(self):
        log.debug("Disconnecting from %s", self._address)
        self._backend.send_command(
            CommandBuilder.connection_disconnect(self._handle))

        self._backend.expect(ResponsePacketType.connection_disconnect)
        log.info("Disconnected from %s", self._address)
        self._handle = None

    @connection_required
    def discover_characteristics(self):
        return self._backend.discover_characteristics(self._handle)




import threading
import traceback

from constants import *
from packet_parser import PacketParser
from utils.sdk_utils import print_and_log, ERROR, EXCEPTION

MyMeData = Enum('MyMeData',
                ['hw_version',
                 'sw_version',
                 'battery_state',
                 'battery_charge_level',
                 'charger_state',
                 'device_time'])


class MyMeDevice(object):
    """
    Class represents an instance of a MyMe device.
    """
    def __init__(self):
        self.handler = None
        self.device_state = None
        # Used to store last state in order to restore it after temporary state
        # change (ie when subscribing)
        self.last_device_state = None
        self.lock = threading.Lock()


        self.hw_version = None
        self.sw_version = None
        self.battery_state = None
        self.battery_charge_level = None
        self.charger_state = None
        self.device_time = None

    def fill_device_info(self):
        """
        Gets all the available device info upon connection and stores it in the
        object. This enables fast response for future user queries regarding
        device info.
        """
        if self.device_state != DeviceState.Connected:
            print_and_log('ERROR: Can\'t fill device info when not connected')

        with self.lock:
            try:
                self.sw_version = PacketParser.parse_version_packet(
                    self.handler.char_read(SOFTWARE_CHARACTERISTIC_UUID))
                print_and_log('got device software_version: {}'
                              ''.format(self.sw_version))
            except Exception as ex:
                print_and_log(EXCEPTION, 'Error getting software vrsion from '
                              'MyMe device\nException Message: {}'.format(ex))
                traceback.print_exc()

            try:
                self.hw_version = PacketParser.parse_version_packet(
                    self.handler.char_read(HARDWARE_CHARACTERISTIC_UUID))
                print_and_log('got device hardware_version: {}'
                              ''.format(self.hw_version))
            except Exception as ex:
                print_and_log(EXCEPTION, 'Error getting harwarde version from '
                              'MyMe Device\nException Message: {}'.format(ex))
                traceback.print_exc()

            try:
                self.device_time =  PacketParser.parse_time_packet(
                    self.handler.char_read(TIME_CHARACTERISTIC_UUID))
                print_and_log('got device time: {}'
                              ''.format(self.device_time))
            except Exception as ex:
                print_and_log(EXCEPTION, 'Error getting device time from '
                              'MyMe Device\nException Message: {}'.format(ex))
                traceback.print_exc()

            try:
                self.battery_charge_level = \
                    PacketParser.parse_battery_charge_level_packet(
                    self.handler.char_read(BATTERY_LEVEL_CHARACTERISTIC_UUID))
                print_and_log('got battery level: {}%'
                              ''.format(self.battery_charge_level))
            except Exception as ex:
                print_and_log(EXCEPTION, 'Error getting battery charge level'
                        'from MyMe Device\nException Message: {}'.format(ex))
                traceback.print_exc()

            try:
                self.charger_state = PacketParser.parse_charger_packet(
                    self.handler.char_read(CHARGER_CHARACTERISTIC_UUID))
                print_and_log('got charger state: {}'
                              ''.format(self.charger_state))
            except Exception as ex:
                print_and_log(EXCEPTION, 'Error getting charger charge '
                        'from MyMe Device\nException Message: {}'.format(ex))
                traceback.print_exc()




__version__="1.0.0"

import cPickle
import datetime
import os
import Queue
import signal
import struct
import time
import traceback

from threading import Thread

from algo.face_session_manager import FaceSessionMananger
from bluetooth.bluetooth_manager import BluetoothManager
from constants import *
from database_manager import *
from delegates import MyMeDelegates
from user_callback_req_types import *
from utils.sdk_utils import db_face_to_myme_face, db_person_to_myme_person
from utils.appdirs import user_data_dir
from myme_device import MyMeData
from myme_data_objects import MyMeFace, MyMePerson


class MyMeSDK(object):
    _CONNECT_TIMEOUT = 7
    _SCAN_TIMEOUT = 5
    _APP_DATA_PATH = user_data_dir(appname=SDK_APPNAME, version=SDK_VERSION)
    _CONFIG_FILE_PATH = os.path.join(_APP_DATA_PATH, 'myme_sdk.config')
    def __init__(self):
        """
        MyMeDelegates class provides an interface for all the delegates that
        the user can redefine. For every delegate that the user wants to redefine,
        he must call the respective register_<delegate_name> function, and pass
        his own redefined delegate function.
        """
        signal.signal(signal.SIGINT, self._sigint_handler)

        self.delegate = MyMeDelegates()
        self.user_callback_thread_stop_event = threading.Event()
        self.user_callback_queue = Queue.Queue()
        self.user_callback_queue_lock = threading.Lock()
        self.user_callback_thread = Thread(target = self._user_callback_thread_entry)
        self.user_callback_thread.start()
        self.is_stopped = False

        # start the BluetoothManager
        self.bt_manager = BluetoothManager(
            sdk_charger_state_updated_cb = self._sdk_charger_state_updated_cb,
            sdk_device_tapped_cb = self._sdk_device_tapped_cb,
            sdk_battery_state_updated_cb = self._sdk_battery_state_updated_cb,
            sdk_battery_charge_level_updated_cb = self._sdk_battery_charge_level_updated_cb,
            sdk_device_state_updated = self._sdk_device_state_updated_cb,
            sdk_scan_complete_cb = self._sdk_scan_complete_cb,
            sdk_connected_to_device_cb = self._sdk_connected_to_device_cb,
            sdk_face_signature_cb = self._sdk_face_signature_cb,
            sdk_face_image_cb = self._sdk_face_image_cb,
            sdk_full_image_cb = self._sdk_full_image_cb,
            sdk_write_to_debug_cb = self._sdk_write_to_debug_cb,
            sdk_visual_topics_cb = self._sdk_visual_topics_cb,
            sdk_ocr_line_cb = self._sdk_ocr_line_cb,
            sdk_barcode_cb = self._sdk_barcode_cb)

        self.bt_manager.start() # start BT manager thread
        self._init_config_folder(self._APP_DATA_PATH)

        self.paired_device_addr = self._get_config_val_from_disk('paired_device_addr')

        if self.paired_device_addr is not None:
            self.bt_manager.set_device_state(DeviceState.Paired)

        # start DatabaseManager
        self.db_manager = DatabaseManager()

        # start FaceSessionMananger
        self.face_session_mananger = FaceSessionMananger(self.db_manager)

    def _user_callback_thread_entry(self):
        """
        This function is the entry point for a dedicated thread whose purpose
        is to call user callbacks (provided by user when calling one of the
        register_* functions). This is needed to prevent deadlock which may be
        caused by the user making SDK API calls from within his delegate
        functions
        """
        while not self.user_callback_thread_stop_event.is_set():
            try:
                item = self.user_callback_queue.get(timeout = 2)
                with self.user_callback_queue_lock:
                    self._handle_user_callback_queue_item(item)
                self.user_callback_queue.task_done()
            except Queue.Empty: # timeout
                continue

    def _handle_user_callback_queue_item(self, req):
        #print_and_log('handling request of type: {}'.format(req.type))
        if req.type == UserCallbackReqType.PairedDeviceStateUpdatedCbReq:
            self.delegate.PairedDeviceStateUpdated(req.new_state)
        elif req.type == UserCallbackReqType.FaceDetectedCbReq:
            self.delegate.FaceDetected(req.face, req.person, req.probability,
                                       req.is_last_face_in_frame)
        elif req.type == UserCallbackReqType.NewPersonCreatedCbReq:
            self.delegate.NewPersonCreated(req.person, req.timestamp)
        elif req.type == UserCallbackReqType.VisualTopicsReceivedCbReq:
            self.delegate.VisualTopicsReceived(req.topics, req.timestamp)
        elif req.type == UserCallbackReqType.OcrLineFoundCbReq:
            self.delegate.OcrLineFound(
                        req.myme_ocr_line, req.timestamp)
        elif req.type == UserCallbackReqType.BarcodeDetectedCbReq:
            self.delegate.BarcodeDetected(req.barcode, req.timestamp)
        elif req.type == UserCallbackReqType.ChargerStateUpdatedCbReq:
            self.delegate.ChargerStateUpdated(req.new_charger_state)
        elif req.type == UserCallbackReqType.DeviceTappedCbReq:
            self.delegate.DeviceTapped()
        elif req.type == UserCallbackReqType.BatteryStateUpdatedCbReq:
            self.delegate.BatteryStateUpdated(req.battery_state)
        elif req.type == UserCallbackReqType.BatteryChargeLevelChangedCbReq:
            self.delegate.BatteryChargeLevelChanged(req.battery_level)

        elif req.type == UserCallbackReqType.ScanCompleteCbReq:
            if req.user_callback is not None:
                req.user_callback(req.sorted_device_list, req.scan_state, req.error)
        elif req.type == UserCallbackReqType.ConnectedToDeviceCbReq:
            if req.user_callback is not None:
                req.user_callback(req.device_addr, req.error)
        elif req.type == UserCallbackReqType.FaceImageReceivedCbReq:
            if req.user_callback is not None:
                print 'calling user callback for new face image yay: {}'.format(req.user_callback)
                req.user_callback(req.myme_face)
        elif req.type == UserCallbackReqType.FullImageReceivedCbReq:
            if req.user_callback is not None:
                req.user_callback(req.myme_image)
        elif req.type == UserCallbackReqType.WriteToDebugCbReq:
            if req.user_callback is not None:
                req.user_callback(req.success)
        else:
            print_and_log('ERROR: GOT REQUEST OF UNKOWN TYPE: {}'.format(req.type))
            print('request data members: {}'.format(req.__dict__))

    def _sigint_handler(self, signum, frame):
        print_and_log('CTRL-C pressed. Will disconnect from device (if '
                      'connected), and exit.')
        self.stop()
        exit(0)

    def __del__(self):
        self.stop()

    def _init_config_folder(self, path):
        if not os.path.exists(path):
            os.makedirs(path)

    # ******************** BLUETOOTH MANAGER CALLBACKS *************************
    # Callback functions provided to the BluetoothManager to notify the SDK
    # when something occurs
    # **************************************************************************

    def _sdk_battery_charge_level_updated_cb(self, battery_level):
        self.user_callback_queue.put(BatteryChargeLevelChangedCbReq(battery_level))

    def _sdk_battery_state_updated_cb(self, new_battery_state):
        self.user_callback_queue.put(BatteryStateUpdatedCbReq(new_battery_state))

    def _sdk_charger_state_updated_cb(self, new_charger_state):
        self.user_callback_queue.put(ChargerStateUpdatedCbReq(new_charger_state))

    def _sdk_device_tapped_cb(self):
        self.user_callback_queue.put(DeviceTappedCbReq())

    def _sdk_device_state_updated_cb(self, new_state):
        self.user_callback_queue.put(PairedDeviceStateUpdatedCbReq(new_state))

    def _sdk_scan_complete_cb(self, sorted_device_list, scan_state, error,
                              user_callback):
        self.user_callback_queue.put(ScanCompleteCbReq(
                sorted_device_list, scan_state, error, user_callback))

    def _sdk_connected_to_device_cb(self, device_addr, error, cb):
        self.user_callback_queue.put(ConnectedToDeviceCbReq(
                device_addr, error, cb))

    def _sdk_visual_topics_cb(self, visual_topics, timestamp):
        self.user_callback_queue.put(VisualTopicsReceivedCbReq(
            topics=visual_topics, timestamp=timestamp))

    def _sdk_ocr_line_cb(self, ocr_line, timestamp, is_frame_end=False):
        self.user_callback_queue.put(OcrLineFoundCbReq(myme_ocr_line=ocr_line,
            timestamp=timestamp))

    def _sdk_barcode_cb(self, barcode, timestamp):
        self.user_callback_queue.put(BarcodeDetectedCbReq(
                barcode=barcode, timestamp=timestamp))

    def _sdk_face_signature_cb(self, detected_face_id, detected_time,
                               signature_int_array, is_last_face_in_frame = False):
        """
        Called by the bt manager whenever a new face signature is received
        Adds new face to the Faces DB

        :param detected_face_id: id of the detected face (application side id)
        :param detected_time: timestamp of when the detection occurred
                              (int seconds from epoch)
        :param signature_int_array: signature of the detected face stored as an
                                    array of integers
        :param: is_last_face_in_frame - True if this is the last face for this frame,
                false otherwise
        :return:
        """
        myme_face, myme_person, probability, new_person_created = \
            self.face_session_mananger.add_face_to_session(
                                    myme_face_id=detected_face_id,
                                    detected_time=detected_time,
                                    signature=signature_int_array)

        self.user_callback_queue.put(FaceDetectedCbReq(
                face=myme_face, person=myme_person, probability=probability,
                is_last_face_in_frame=is_last_face_in_frame))
        if new_person_created:
            self.user_callback_queue.put(NewPersonCreatedCbReq(
                    person=myme_person, timestamp=time.time()))


    def _sdk_face_image_cb(self, face_db_id, detected_time,
                           image_bytearray, user_callback):
        """
        Stores received image into the Faces DB with myme_face.db_id being the key
        :param face_myme_id: Id of face in the SDK DB for which image was fetched
        :param image_bytearray: Fetched image stored as a bytearray
        """
        self.db_manager.add_face_image_to_id(face_db_id, image_bytearray)
        myme_face = MyMeFace(id=face_db_id,
                         detected_time=detected_time,
                         face_image=bytearray_to_image(image_bytearray))
        self.user_callback_queue.put(FaceImageReceivedCbReq(
                myme_face, user_callback))

    def _sdk_full_image_cb(self, myme_image, user_callback):
        self.user_callback_queue.put(FullImageReceivedCbReq(
                myme_image, user_callback))

    def _sdk_write_to_debug_cb(self, success, user_callback):
        self.user_callback_queue.put(WriteToDebugCbReq(success, user_callback))

    # **************************************************************************
    # ************************* CONNECT, SCAN & PAIR ***************************
    # **************************************************************************
    def quick_connect(self):
        """
        This method scans for devices, pairs to the device that has the strongest
        signal and connects to it. The completion block is called when the
        device finishes connecting or connection attempt fails
        """
        self.bt_manager.scan_for_peripherals_with_timeout(
                self._SCAN_TIMEOUT, True, self._cb_quick_connect_scan)

    def connect_to_paired_device_with_timeout(self, timeout_secs=_CONNECT_TIMEOUT,
                                              cb=None, block = False):
        """
        Connects to the paired device, In case connection fails, tries to
        reconnect until success or until <timeout_secs> pass.

        :param timeout: timeout time for connection request in seconds
        :param cb: (optional) callback function
        :param block: if True -blocks the caller until done
        """
        if self.paired_device_addr is None:
            error = 'ERROR: You must first pair a device before connecting.'
            self._sdk_connected_to_device_cb(None, error, cb)
        else:
            self.bt_manager.connect_to_paired_device_with_timeout(
                    timeout_secs, self.paired_device_addr, cb, block)


    def disconnect_from_device(self, callback=None):
        """
        Disconnect from the device

        :param callback:
        """
        self.bt_manager.disconnect_from_device(callback)

    # TODO - add excpetion - cb must be suppiled when block is false
    # TODO - consider remove block or at least True by default
    def scan_for_devices_with_timeout(self, timeout_sec, block=True, cb=None):
        """
        This method scans for devices until <timeout_sec> pass.
        When done the ScanComplete delegate is called with the full list of detected devices

        :param timeout_sec: timeout in seconds
        :param block: if True, blocks until scan is done
        :param cb: Optional user callback function (if None, default is be used)
        """
        self.bt_manager.scan_for_peripherals_with_timeout(timeout_sec, block, cb)


    def _cb_quick_connect_scan(self, device_list, scan_state,
                               error=None, subscribe_to_all=True):
        """
        Calledback by bluetooth manager after a scan that was initiated by
        quick_connect

        :param device_list: List of devices.
        Each device is represented by dict with the following keys:
            "address" (device's MAC address)
            "name" (device name)
            "rssi" (signal strength)
        :param scan_result: scan compeltion reason (ScanState)
        :param subscribe_to_all: If true, device will subscribe to notifications
            from all available services
        :param error: in case of error, this string contains the error message
        :return:
        """
        if error:
            print error
        if device_list == []:
            print 'No devices found'
        else:
            self.pair_device(device_list[0]['address'])
            self.connect_to_paired_device_with_timeout(timeout_secs=7,
                                                       block=True)

    def _get_config_val_from_disk(self, key):
        """
        Fetches config value assigned to <key> from the app config file on disk

        :param key: key for the value to return from the config
        :return: value assigned to <key>, None if <key> not in config
        """
        if not os.path.exists(self._CONFIG_FILE_PATH):
            return None

        # NOTE: MUST OPEN PICKLED FILE WITH "b" MODE FOR WINDOWS COMPATABILITY
        with open(self._CONFIG_FILE_PATH, 'rb') as config_pickle:
            try:
                config_dict = cPickle.load(config_pickle)
                return config_dict.get(key, None)
            except Exception as ex:
                print "Didn't find stored config."
                return None


    def _store_config_val_to_disk(self, key, val):
        """
        Stores a new value into the application config file under <key>

        :param key: key for the value to store in the config
        :param val: value to store
        """
        # NOTE: MUST OPEN PICKLED FILE WITH "b" MODE FOR WINDOWS COMPATABILI    TY
        config_dict = {}
        if os.path.exists(self._CONFIG_FILE_PATH):
            with open(self._CONFIG_FILE_PATH, 'rb') as config_file:
                try:
                    config_dict = cPickle.load(config_file)
                except Exception as ex:
                    print "Didn't find stored config."
                    print 'Exception msg: {}'.format(ex)
                    traceback.print_exc()


        config_dict[key]=val
        with open(self._CONFIG_FILE_PATH, 'wb') as config_file:
            cPickle.dump(config_dict, config_file)

    def pair_device(self, device_addr):
        """
        This method saves the paired <device_addr> for future connection.
        Pairing must be done before connection.

        :param peripheral: Id the ID of the device to pair.
        :return: True if device_addr successfully stored, False otherwise
        """
        self.paired_device_addr = device_addr
        self.bt_manager.set_device_state(DeviceState.Paired)
        self._store_config_val_to_disk('paired_device_addr', device_addr)

    def clear_paired_device(self):
        self.paired_device_addr = None
        # TODO - why is this commented out?
        #self._store_config_val_to_disk('paired_device_addr', None)


    def get_paired_device_address(self):
        """
        :return: String containing the device address that was saved when pair_device() was called. Or None if no device has been paired
        """
        return self.paired_device_addr

    def get_device_connectivity_state(self):
        return self.bt_manager.get_device_state()

    # **************** MyMe Device Interaction Functions ***********************
    # Functions that request or send to various data from from/to the device
    # **************************************************************************

    # TODO - comment out
    def _write_to_debug(self, text_to_write, callback=None):
        """
         This method enables writing text to the device.
         :param text_to_write The text to be written to the device
         :param callback: optional callback function
        """
        self.bt_manager.write_to_debug(text_to_write, callback)

    def get_battery_charge_level(self, battery_charge_level_cb):
        """
        Gets the percent charge of the battery from the device

        """
        self.bt_manager.get_info_from_device(MyMeData.battery_charge_level,
                                             battery_charge_level_cb)

    def get_battery_state(self, battery_state_cb):
        """
        Get the BatteryState of the device battery. The possible values are:
                BatteryOK = 0
                BatteryCharging = 1
                BatteryLow = 2
                BatteryCritical = 3 # Device will shutdown upon reaching state

        :param battery_state_cb: callback function
        """
        self.bt_manager.get_info_from_device(MyMeData.battery_state,
                                             battery_state_cb)


    def get_charger_state(self, charger_state_cb):
        """
        Reads the charger state
        """
        self.bt_manager.get_info_from_device(MyMeData.charger_state,
                                             charger_state_cb)


    def get_device_time(self, device_time_cb):
        """
        Reads the time set on the connected device
        """
        self.bt_manager.get_info_from_device(MyMeData.device_time,
                                             device_time_cb)

    def get_device_hardware_version(self):
        return self.bt_manager.get_device_hardware_version()

    def get_device_software_version(self):
        return self.bt_manager.get_device_software_version()

    def _request_full_image(self, frame_idx, full_image_cb):
        """

        :param frame_idx: Index of frame for which to request a full image
        :param full_image_cb: callback function to be called after full image
                              request is handled
        """
        self.bt_manager.request_full_image(frame_idx, full_image_cb)


    # ******************* DATABASE INTERACTION FUNCTIONS ***********************
    # Functions that send or request data from the SDK database
    # **************************************************************************

    def get_all_people(self):

        """
        Returns a list with all the people that exist in the DB

        :return list of MyMePerson
        """
        return [db_person_to_myme_person(person) for person in
                self.db_manager.get_all_people()]

    # TODO - person - pass by id
    def set_name_to_person(self, my_me_person, name):
        """
        Sets the given name to the given my_me_person

        :param my_me_person: my_me_person object
        :param name: the nsame to set
        :return: None
        """
        db_person = self.db_manager.get_person_by_person_id(my_me_person.db_id)
        if db_person is None:
             print_and_log('ERROR: person id {0} does not exist'.
                           format(my_me_person.db_id), ERROR)
             return
        db_person.name = name
        self.db_manager.commit_db()

    # TODO - person - pass by id
    def set_profile_picture(self, my_me_person, profile_pic):
        """
        Sets the given profile picture to the given my_me_person.

        :param my_me_person: my_me_person object
        :param profile_pic: the prfoile picture to set
        :return: None
        """
        db_person = self.db_manager.get_person_by_person_id(my_me_person.db_id)
        db_person.profile_pic = profile_pic
        self.db_manager.commit_db()

    def get_person(self, person_id):
        """
        Fetches a person by Id from the DB, with his list of faces

        :param person_id: the person Id to fetch
        :return: MyMePerson or None if not found
        """
        return db_person_to_myme_person(
                (self.db_manager.get_person_by_person_id(person_id)))

    def remove_person(self, person_id):
        """
        Removes the given person from the DB,
         any person face(s) is(are) deleted as well

        :param person_id:
        :return:
        """
        db_person = self.db_manager.get_person_by_person_id(person_id)
        if db_person is not None:
            self.db_manager.remove(db_person)

    def associate_face(self, myme_face, myme_person):
        """
        Associates the given MyMeface with the given MyMePerson
        Note the preivouse association of the given face is removed from the relevant face

        :param myme_face: the face to associate
        :param myme_person: the person to associate with
        :return: None
        """
        db_person = self.db_manager.get_person_by_person_id(myme_person.db_id)
        if db_person is None:
            print_and_log('ERROR: person id {} does not exist'.
                          format(myme_person.db_id), ERROR)
            return
        db_face   = self.db_manager.get_face_by_face_id(myme_face.id)
        if db_face is None:
            print_and_log('ERROR: db face id {0} does not exist'.
                          format(myme_face.id), ERROR)
            return

        db_person.faces.append(db_face)
        self.db_manager.commit_db()

    def delete_face(self, face_id):
        """
        Deletes the given face from the DB
        The reference to the face from the relevant person is also deleted

        :param face_id: the face id to delete
        :return: None
        """
        db_face = self.db_manager.get_face_by_face_id(face_id)
        self.db_manager.remove(db_face)

    def get_all_faces_from_time(self, start_time, end_time):
        """
        Returns all MyMeFaces detected between start_time and end_time

        :param start_time: state time (DateTime object)
        :param end_time: end time (DateTime object)
        :return: list of faces matching the query
        """
        db_faces = self.db_manager.get_all_faces_from_time(start_time, end_time)
        return [db_face_to_myme_face(face) for face in db_faces]

    def get_all_faces(self):
        """
        Returns a list of all MyMeFace objects in the DB

        :return: list of all MyMeFace objects in the DB
        """
        return [db_face_to_myme_face(face) for face in
                self.db_manager.get_all_faces()]

    def request_image_for_face_id(self, db_face_id, face_image_cb):
        """
        Requests the image from the device that is associated with db_id.
        If the image arrives, it will be added automatically to the face DB.

        :param db_face_id: ID of the face in Faces DB
        :param face_image_cb: Callback function that receives the fetched face and optionally an error message
        :return:
        """
        db_face = self.db_manager.get_face_by_face_id(db_face_id)

        self.bt_manager.request_face_image(
                db_face_id=db_face_id,
                myme_face_id=db_face.myme_id,
                face_image_cb=face_image_cb)

    def register_paired_device_state_updated_cb(self, cb):
        self.delegate.PairedDeviceStateUpdated = cb

    def register_face_detected_cb(self, cb):
        self.delegate.FaceDetected = cb

    def register_new_person_created_cb(self, cb):
        self.delegate.NewPersonCreated = cb

    def _register_visual_topics_received_cb(self, cb):
        self.delegate.VisualTopicsReceived = cb

    def register_ocr_line_found_cb(self, cb):
        self.delegate.OcrLineFound = cb

    def register_device_tapped_cb(self, cb):
        self.delegate.DeviceTapped = cb

    def register_barcode_detected_cb(self, cb):
        self.delegate.BarcodeDetected = cb

    def stop(self):
        if not self.is_stopped:
            # stop the bluetooth manager, this would also disconnect from device and stop bgapi lib
            self.bt_manager.stop()

            # stop the user callback thread
            self.user_callback_thread_stop_event.set()
            self.user_callback_thread.join()

            # stop the user callback thread
            self.face_session_mananger.stop()

            self.is_stopped = True

    def _set_device_to_current_time(self):
        """
        TODO
        We should not expose the ability to set the device time to the user,
        it's internal, and in any case the SDK sets it to current time whenever
        it's connected and the time is off, so it would just cause confusion in
        cases the user sets the API to a different time.
        """
        raise NotImplementedError

    """
    /**
     Returns the person that has the best match to the specified face that the matching score is under the given score. (lower score means better match).
     @param face The MyMeFace that will be used to find the best match.
     @param score The score that the matching Person has to be under. we recommend to insert 2.5.
     @return The best matched MyMePerson, returns nil if no such person exists.
     */
    -(MyMePerson*_Nullable) bestPersonMatchForFace:(nonnull MyMeFace*)face underScore:(CGFloat)score;

    /**
     Calculates the matching score for all people in the DB to the given Face. (lower score means better match).
     @param face The face you want to detect possible matches.
     @param underScore People that there score is lower than this parameter will not be sent back. if you want to get all people insert here FLT_MAX.
     @return An array with Objects of type MatchedPersonData, each MatchedPersonData Object contains the Person ID and the matching score. if no matches found, return nil.
     @see MatchingData.h class
     */
    -(NSArray*_Nullable) detectPossibleMatchesFrom:(nonnull MyMeFace*)face underScore:(CGFloat)underScore;

    /**
     Calculates the similarity score of 2 Faces, as lower the score is the chances of a match are higher.
     @param face1 The first face you want to calculate similarity.
     @param face2 The second face you want to calculate similarity.
     @return CGFloat The matching score.
     */
    -(CGFloat) similarityOf:(nonnull MyMeFace*)face1 and:(nonnull MyMeFace*)face2;
    """



import sys
import threading
import time

from ..database_manager import *
from face_session import FaceSession
from ..utils.sdk_utils import db_person_to_myme_person

class FaceSessionMananger():

    THREAD_WAIT_TIME_SEC = 5
    SIGANTURE_LEN        = 128
    SESSION_MATCHING_THRESHOLD = 25.65
    MAX_SESSION_SIZE = 50
    MAX_IDLE_SEC = 60

    def __init__(self, db_manager):
        """
        Initializes the FaceSessionMananger.
        :param session_update_cb: called whenever the session is updated (new face detected)
        :param session_eneded_cb: called whenever the session has ended
        :return:
        """
        self.db_manager = db_manager

        self.open_sessions = [] # list of FaceSession

    def stop(self):
        """
        Closes all sessions
        """
        for session in self.open_sessions:
            session.close()

    def close_expired_sessions(self):
        for i, session in enumerate(self.open_sessions):
            if time.time() - session.last_face_added_timestamp > self.MAX_IDLE_SEC:
                session.close()
                del self.open_sessions[i]

    def find_best_session_for_sig(self, signature):
        """
        Finds and returns the session in open_sessions with the signature
        closest to the current signature.
        Or, if no session with a close enough signature exists, creates a new
        session
        :param signature: Face signature (array of ints)
        :return: index of session into which signature should be inserted
        """
        curr_session_idx = None

        # go over list of sessiosn
        min_dist_to_session = sys.float_info.max
        min_session_idx = None
        for i, session in enumerate(self.open_sessions):
            # get average distance of the signature to the session
            cur_distance_to_session = session.get_sig_to_session_score(signature)
            if  cur_distance_to_session < min_dist_to_session:
                min_dist_to_session = cur_distance_to_session
                min_session_idx = i

        # use the closest one if it's close enough (smaller than SESSION_MATCHING_THRESHOLD)
        if min_dist_to_session <= self.SESSION_MATCHING_THRESHOLD:
            print 'Found matching session! (dist: {})'.format(min_dist_to_session)
            return min_session_idx

        # otherwise open a new session
        print ('No matching session found (dist to closest: {}). Opening a new '
               'session.'.format(min_dist_to_session))
        self.open_sessions.append(FaceSession(self.db_manager))
        min_session_idx = len(self.open_sessions) - 1
        return min_session_idx


    def add_face_to_session(self, myme_face_id, detected_time, signature):
        """
        Update the current session with a newly detected face.
        If face is added to a session that has >= 3 faces and has no
        matching person, a new person is created containing the session's faces
        and the session is closed. <new_person_created> is True in such case.
        :param myme_face_id: Id of the face internal to MyMe application
                             will be stored in the database to enable later
                             fetching of image for the face
        :param detected_time: Time when face was deteced by MyMe
        :param signature: Signature of the detected faces (array of 128 ints)
        """
        # 0. Close all sessions that haven't been updated in more than MAX_IDLE_SEC
        self.close_expired_sessions()

        # 1. find best session (this might open a new session if needed)
        curr_session_idx = self.find_best_session_for_sig(signature)
        curr_session = self.open_sessions[curr_session_idx]

        # 2. add face to the best session and return a MyMeFace object holding the face
        myme_face = curr_session.add_face(myme_face_id, detected_time, signature)

        # 3. find person most closely matching curr_session or None if not found w/in a threshold
        matching_db_person, probability = curr_session.find_best_person_for_sigs()
        if matching_db_person is not None:
            print 'found best person for current session, person id: ' \
                  '{}, probablity of match: {}%' \
                  ''.format(matching_db_person.id, probability)
        else:
            print 'didn\'t find any people matching current session'

        # 4. Close session if needed (will create new person if no matching person exist already)
        new_person_created = False
        if (matching_db_person is None and
            len(curr_session.db_face_ids) >= FaceSession.MIN_SIZE_TO_CREATE_NEW_PERSON):
            print 'creating new person for current session and closing session'
            curr_session.create_person_and_close()
            new_person_created = True
            matching_db_person = self.db_manager.get_person_by_person_id(curr_session.db_person_id)
            probability = curr_session.db_probability
            del self.open_sessions[curr_session_idx]
        elif len(curr_session.db_face_ids) >= self.MAX_SESSION_SIZE:
            curr_session.close()
            del self.open_sessions[curr_session_idx]


        # 5. Return
        myme_person = None
        if matching_db_person is not None:
            myme_person = db_person_to_myme_person(matching_db_person)
        return myme_face, myme_person, probability, new_person_created





import binascii
import io
import logging

from PIL import Image

from ..myme_data_objects import MyMeFace, MyMePerson

DEBUG = 'debug'
ERROR = 'error'
EXCEPTION = 'exception'

class colors(object):
    # HIGH INTENSITY COLORS
    CYAN =   '\033[96m'
    PURPLE = '\033[95m'
    BLUE =   '\033[94m'
    YELLOW = '\033[93m'
    GREEN =  '\033[92m'
    RED =    '\033[91m'

    # REGULAR COLORS
    black =     '\033[30m'
    red =       '\033[31m'
    green =     '\033[32m'
    yellow =    '\033[33m'
    blue =      '\033[34m'
    purple =    '\033[35m'
    cyan =      '\033[36m'
    lightgrey = '\033[37m'

    # MARKINGS
    BOLD =      '\033[1m'
    UNDERLINE = '\033[4m'
    END =           '\033[0m'
    REVERSE =       '\033[07m'
    STRIKETHROUGH = '\033[09m'
    INVISIBLE =     '\033[08m'
    DISABLE =       '\033[02m'


def print_and_log(message, type = 'info', bold = False, color = None,
                  logger_name = None, comma=False, use_console_only=False):
    if logger_name:
        main_logger = logging.getLogger(logger_name)
    else:
        main_logger = logging.getLogger()

    if (type == ERROR or type == EXCEPTION) and color == None:
        bold = True
        color = colors.RED

    if bold and color:
        print color  + colors.BOLD + message + colors.END
    elif color:
        print color + message + colors.END
    elif bold:
        print colors.BOLD + message + colors.END
    else:
        print message
    if use_console_only:
        return
    if (type == 'info'):
        main_logger.info(message)
    elif (type == EXCEPTION):
        main_logger.exception(message)
    elif (type == ERROR):
        main_logger.error(message)
    elif type == DEBUG:
        main_logger.debug(message)

1
def int2bytes(i):
    hex_string = '%x' % i
    n = len(hex_string)
    return binascii.unhexlify(hex_string.zfill(n + (n & 1)))

def int_to_bytes(num, length=0, little_endian=True):
    """
    Given an int, returns a byte array representing the int
    :param num: The number to convert
    :param length: Length of output string (will zero-pad to get to this length)
    :param little_endian: if True, will return byte array in reversed order
    :return:
    """
    if not length:
        length = len(str(num))
    h = '{:02X}'.format(num)
    s = ('0'*(len(h) % 2) + h).zfill(length*2).decode('hex')
    return s[::-1] if little_endian else s

def db_face_to_myme_face(db_face):
    return MyMeFace(id=db_face.id, detected_time=db_face.detected_time,
                    signature=db_face.get_signature(),
                    face_image=db_face.get_image(),
                    image_fetch_failed=db_face.image_fetch_failed)

def db_person_to_myme_person(db_person):
    if db_person is None:
        return None
    myme_person = MyMePerson(db_id=db_person.id,
                        name=db_person.name,
                        profile_pic=db_person.profile_pic,
                        faces=[db_face_to_myme_face(face) for
                                   face in db_person.faces])
    return myme_person

def bytearray_to_image(image_bytearray):
    if image_bytearray is None:
        return None
    image = Image.open(io.BytesIO(image_bytearray))
    return image